# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import typing as T

from .common import Compiler
from ..utils import get_empty_file


_powermake_warning_flags_to_msvc_flags: T.Dict[str, T.List[str]] = {
    "-w": ["/W0"],
    "-Wall": ["/W3"],
    "-Wextra": ["/W4"],
    "-Weverything": ["/Wall"],
    "-Wsecurity": ["/Wall"],
    "-Werror": ["/WX"],
    "-Wpedantic": [],
    "-pedantic": [],
    "-Wswitch": ["/we4062"],
    "-Wswitch-enum": ["/we4061"],
    "-fanalyzer": ["/analyze"],
    "-ffuzzer": ["/fsanitize=address", "/fsanitize=fuzzer"]
}

_powermake_optimization_flags_to_msvc_flags: T.Dict[str, T.List[str]] = {
    "-O0": ["/Od"],
    "-Og": ["/Od"],
    "-O": ["/O1"],
    "-O1": ["/O1"],
    "-O2": ["/O2"],
    "-O3": ["/Ox", "/fp:fast"],
    "-Os": ["/O1", "/GL"],
    "-Oz": ["/O1", "/Oi", "/GL"],
    "-Ofast": ["/Ox", "/fp:fast"],
    "-fomit-frame-pointer": ["/Oy"],
}

_powermake_machine_flags_to_msvc_flags: T.Dict[str, T.List[str]] = {
    "-m32": [],
    "-m64": [],
    "-march=native": [],
    "-mtune=native": [],
    "-mmmx": ["/arch:MMX"],
    "-msse": ["/arch:SSE"],
    "-msse2": ["/arch:SSE2"],
    "-msse3": ["/arch:SSE3"],
    "-mavx": ["/arch:AVX"],
    "-mavx2": ["/arch:AVX2"]
}

_powermake_flags_to_msvc_flags: T.Dict[str, T.List[str]] = {
    **_powermake_warning_flags_to_msvc_flags,
    **_powermake_optimization_flags_to_msvc_flags,
    **_powermake_machine_flags_to_msvc_flags,
    "-g": ["/Z7"],
    "-fdiagnostics-color": [],
    "-fsecurity=1": [],
    "-fsecurity=2": [],
    "-fsecurity": []
}


class CompilerMSVC(Compiler):
    type: T.ClassVar = "msvc"
    default_path: T.ClassVar = "cl"
    obj_extension: T.ClassVar = ".obj"
    translation_dict: T.ClassVar = _powermake_flags_to_msvc_flags

    def __init__(self, path: str = "cl") -> None:
        super().__init__(path)

    def format_args(self, defines: T.List[str], includedirs: T.List[str], flags: T.List[str] = []) -> T.List[str]:
        return [f"/D{define}" for define in defines] + [f"/I{includedir}" for includedir in includedirs] + self.translate_flags(flags)

    def basic_compile_command(self, outputfile: str, inputfile: str, args: T.List[str] = []) -> T.List[str]:
        return [self.path, "/c", "/nologo", "/Fo" + outputfile, inputfile, *args]

    def check_if_arg_exists(self, arg: str) -> bool:
        return subprocess.run([self.path, arg, "/E", "/options:strict", get_empty_file()], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL).returncode == 0

class CompilerClang_CL(CompilerMSVC):
    type: T.ClassVar = "clang-cl"
    default_path: T.ClassVar = "clang-cl"

    def __init__(self, path: str = "clang-cl") -> None:
        super().__init__(path)



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "SharedLinker", "SharedLinkerGNU", "SharedLinkerGCC", "SharedLinkerClang", "SharedLinkerGPlusPlus", "SharedLinkerClangPlusPlus", "SharedLinkerMSVC", "SharedLinkerClang_CL"
]

import typing as T

from .common import SharedLinker
from .gnu import SharedLinkerGNU, SharedLinkerGCC, SharedLinkerClang, SharedLinkerGPlusPlus, SharedLinkerClangPlusPlus
from .msvc import SharedLinkerMSVC, SharedLinkerClang_CL


_shared_linker_types: T.Dict[str, T.Callable[[], SharedLinker]] = {
    "gnu": SharedLinkerGNU,
    "gcc": SharedLinkerGCC,
    "g++": SharedLinkerGPlusPlus,
    "clang": SharedLinkerClang,
    "clang++": SharedLinkerClangPlusPlus,
    "msvc": SharedLinkerMSVC,
    "clang-cl": SharedLinkerClang_CL
}


def GenericSharedLinker(shared_linker_type: str) -> T.Union[T.Callable[[], SharedLinker], None]:
    if shared_linker_type not in _shared_linker_types:
        return None
    return _shared_linker_types[shared_linker_type]


def get_all_shared_linker_types() -> T.Set[str]:
    return set(_shared_linker_types.keys())



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
import typing as T

from ..tools import Tool


class SharedLinker(Tool, abc.ABC):
    shared_lib_extension: T.ClassVar = ""

    def __init__(self, path: str) -> None:
        Tool.__init__(self, path)

    @abc.abstractmethod
    def format_args(self, shared_libs: T.List[str], flags: T.List[str]) -> T.List[str]:
        return []

    @abc.abstractmethod
    def basic_link_command(self, outputfile: str, objectfiles: T.Iterable[str], archives: T.List[str] = [], args: T.List[str] = []) -> T.List[str]:
        return []



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import typing as T

from ..tools import _powermake_flags_to_gnu_flags
from .common import Compiler
from ..utils import get_empty_file


class CompilerGNU(Compiler):
    type: T.ClassVar = "gnu"
    default_path: T.ClassVar = "cc"
    obj_extension: T.ClassVar = ".o"

    def __init__(self, path: str = "cc"):
        super().__init__(path)

    def format_args(self, defines: T.List[str], includedirs: T.List[str], flags: T.List[str] = []) -> T.List[str]:
        return [f"-D{define}" for define in defines] + [f"-I{includedir}" for includedir in includedirs] + self.translate_flags(flags)

    def basic_compile_command(self, outputfile: str, inputfile: str, args: T.List[str] = []) -> T.List[str]:
        return [self.path, "-c", "-o", outputfile, inputfile, *args]

    def check_if_arg_exists(self, arg: str) -> bool:
        return subprocess.run([self.path, arg, "-E", "-x", "c", get_empty_file()], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL).returncode == 0


class CompilerGNUPlusPlus(CompilerGNU):
    type: T.ClassVar = "gnu++"
    default_path: T.ClassVar = "c++"
    translation_dict = _powermake_flags_to_gnu_flags

    def __init__(self, path: str = "c++"):
        super().__init__(path)
        not_supported = {'-Wjump-misses-init', '-Wmissing-prototypes', '-Wmissing-variable-declarations', '-Wnested-externs', '-Wstrict-prototypes', '-Wunsuffixed-float-constants'}
        self.translation_dict["-Weverything"] = list(filter(lambda x: x not in not_supported, self.translation_dict["-Weverything"]))

    def check_if_arg_exists(self, arg: str) -> bool:
        return subprocess.run([self.path, arg, "-E", "-x", "c++", get_empty_file()], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL).returncode == 0


class CompilerGCC(CompilerGNU):
    type: T.ClassVar = "gcc"
    default_path: T.ClassVar = "gcc"

    def __init__(self, path: str = "gcc"):
        super().__init__(path)


class CompilerGPlusPlus(CompilerGNUPlusPlus):
    type: T.ClassVar = "g++"
    default_path: T.ClassVar = "g++"

    def __init__(self, path: str = "g++"):
        super().__init__(path)


class CompilerClang(CompilerGNU):
    type: T.ClassVar = "clang"
    default_path: T.ClassVar = "clang"

    def __init__(self, path: str = "clang"):
        super().__init__(path)
    
    def check_if_arg_exists(self, arg: str) -> bool:
        return subprocess.run([self.path, arg, "-Werror=unknown-warning-option", "-Werror=unused-command-line-argument", "-E", "-x", "c", get_empty_file()], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL).returncode == 0


class CompilerClangPlusPlus(CompilerGNUPlusPlus):
    type: T.ClassVar = "clang++"
    default_path: T.ClassVar = "clang++"

    def __init__(self, path: str = "clang++"):
        super().__init__(path)
    
    def check_if_arg_exists(self, arg: str) -> bool:
        return subprocess.run([self.path, arg, "-Werror=unknown-warning-option", "-Werror=unused-command-line-argument", "-E", "-x", "c++", get_empty_file()], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL).returncode == 0


# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import typing as T

from .common import Archiver


class ArchiverMSVC(Archiver):
    type: T.ClassVar = "msvc"
    static_lib_extension: T.ClassVar = ".lib"

    def __init__(self, path: str = "lib"):
        super().__init__(path)

    def basic_archive_command(self, outputfile: str, inputfiles: T.Iterable[str], args: T.List[str] = []) -> T.List[str]:
        return [self.path, "/nologo", *args, "/out:" + outputfile, *inputfiles]
    
    def check_if_arg_exists(self, arg: str) -> bool:
        return subprocess.run([self.path, arg, "/WX"], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL).returncode == 0



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys
import subprocess
import typing as T
import __main__ as __makefile__
from threading import Lock

from .config import Config
from .display import print_info, print_debug_info


_print_lock = Lock()
_commands_counter = 0


def run_command(config: Config, command: T.Union[T.List[str], str], shell: bool = False, target: T.Union[str, None] = None, output_filter: T.Union[T.Callable[[bytes], bytes], None] = None, **kwargs: T.Any) -> int:
    global _commands_counter
    
    returncode = 0
    try:
        output = subprocess.check_output(command, shell=shell, stderr=subprocess.STDOUT, **kwargs)
    except subprocess.CalledProcessError as e:
        output = e.output
        returncode = e.returncode
    
    if output_filter is not None:
        output = output_filter(output)

    _print_lock.acquire()

    _commands_counter += 1

    print_info(f"Generating {os.path.basename(target)}" if target is not None else "", config.verbosity, _commands_counter, config.nb_total_operations)

    print_debug_info(command, config.verbosity)

    sys.stdout.flush()
    stdout_fd = sys.stdout.fileno()
    os.write(stdout_fd, output)

    _print_lock.release()

    return returncode


def resolve_path(current_folder: str, additional_includedirs: T.List[str], filepath: str) -> T.Union[str, None]:
    path = os.path.join(current_folder, filepath)
    if os.path.exists(path):
        return path
    for folder in additional_includedirs:
        path = os.path.join(folder, filepath)
        if os.path.exists(path):
            return path
    return None


def is_file_uptodate_recursive(output_date: float, filename: str, additional_includedirs: T.List[str], headers_already_found: T.List[str] = []) -> bool:
    try:
        if os.path.getmtime(filename) >= output_date:
            return False
    except OSError:
        return False

    if not filename.endswith((".c", ".cpp", ".cc", ".C", ".h", ".hpp", ".lex", ".y")):
        return True

    headers_found = []
    file = open(filename, "r", encoding="latin1")

    # The four lines under this comment are here for the compatibility with python < 3.8
    # They are equivalent to `while line := file.readline():`
    while True:
        line = file.readline()
        if not line:
            break

        i = 0
        while i < len(line) and (line[i] == ' ' or line[i] == '\t'):
            i += 1
        if i >= len(line):
            continue
        if line[i] != '#':
            continue
        i += 1
        while i < len(line) and (line[i] == ' ' or line[i] == '\t'):
            i += 1
        if i >= len(line):
            continue
        if line[i:].startswith("include"):
            i += len("include")
            if i >= len(line) or (line[i] != ' ' and line[i] != '\t'):
                continue
            i += 1
            while i < len(line) and (line[i] == ' ' or line[i] == '\t'):
                i += 1
            if i >= len(line) or line[i] != '"':
                continue
            i += 1
            j = i
            while j < len(line) and line[j] != '"':
                j += 1
            if j >= len(line):
                continue

            if line[i:j] not in headers_found:
                headers_found.append(line[i:j])

    file.close()

    if len(headers_found) == 0:
        return True

    current_folder = os.path.dirname(filename)
    new_paths = []
    for i in range(len(headers_found)):
        path = resolve_path(current_folder, additional_includedirs, headers_found[i])
        if path is not None and path not in headers_already_found:
            headers_already_found.append(path)
            new_paths.append(path)

    del headers_found

    for path in new_paths:
        if not is_file_uptodate_recursive(output_date, path, additional_includedirs, headers_already_found):
            return False

    return True


def needs_update(outputfile: str, dependencies: T.Iterable[str], additional_includedirs: T.List[str]) -> bool:
    """Returns whether or not `outputfile` is up to date with all his dependencies

    If `dependencies` includes C/C++ files and headers, all headers these files include recursively will be add as hidden dependencies.

    Args:
        outputfile (str): the path to the target file
        dependencies (set): a set of all files on which `outputfile` depends
        additional_includedirs (list): The list of additional includedirs used by the compiler. This is necessary to discover hidden dependencies.

    Returns:
        bool: True if `outputfile` is **not** up to date with all his dependencies and hidden dependencies.
    """
    try:
        output_date = os.path.getmtime(outputfile)
    except OSError:
        return True
    
    if hasattr(__makefile__, '__file__') and os.path.getmtime(__makefile__.__file__) >= output_date:
        return True
    
    headers_already_found: T.List[str] = []
    for dep in dependencies:
        if not is_file_uptodate_recursive(output_date, dep, additional_includedirs, headers_already_found):
            return True

    return False


class Operation:
    def __init__(self, outputfile: str, dependencies: T.Iterable[str], config: Config, command: T.List[str]):
        """Provide a simple object that can execute a command only if it's needed.

        Args:
            outputfile (str): the path to the target file
            dependencies (set): a set of all files on which `outputfile` depends
            config (Config): A powermake.Config object, the additional_includedirs in it should be completed
            command (list): The command that will be executed by subprocess. It's a list representing the argv that will be passed to the program at the first list position.
        """
        self.outputfile = outputfile
        self.dependencies = dependencies
        self._hidden_dependencies = None
        self.command = command
        self.config = config

    def execute(self, force: bool = False) -> str:
        global _commands_counter

        """Verify if the outputfile is up to date with his dependencies and if not, execute the command.

        Args:
            force (bool, optional): If True, this function will always execute the command without verifying if this is needed.

        Raises:
            RuntimeError: If the command fails.

        Returns:
            str: The outputfile, like that we can easily parallelize this method.
        """
        if force or needs_update(self.outputfile, self.dependencies, self.config.additional_includedirs):

            if run_command(self.config, self.command, target=self.outputfile) == 0:
                return self.outputfile
            else:
                raise RuntimeError(f"Unable to generate {os.path.basename(self.outputfile)}")
        else:
            _print_lock.acquire()
            _commands_counter += 1
            _print_lock.release()
        return self.outputfile

    def get_json_command(self) -> T.Dict[str, T.Any]:
        json_command = {
            "directory": os.getcwd(),
            "arguments": self.command,
            "output": self.outputfile
        }
        deps = list(self.dependencies)
        if len(deps) > 0:
            json_command["file"] = deps[0]

        return json_command



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def simplify_architecture(architecture: str) -> str:
    arch = architecture.lower()
    if arch in ["x86", "x32", "80x86", "8086", "80386", "i286", "i386", "i486", "i586", "i686", "i786", "amd386", "am386", "amd486", "am486", "amd-k5", "amd-k6", "amd-k7"]:
        return "x86"

    if arch in ["x86_64", "x86-64", "x64", "amd64", "intel64"]:
        return "x64"

    if arch in ["arm32", "arm", "armeabi", "armv4", "armv4t", "armv5", "armv5t", "armv5te", "armv6", "armv6-m", "armv6j", "armv6k", "armv6kz", "armv6t2", "armv6z", "armv6zk", "armv6s-m", "armv7", "armv7a", "armv7s", "armv7m", "armv7r", "armv7l", "armv7-a", "armv7-m", "armv7-r", "armeabi-v7a", "armv7ve", "armv7-r", "armv7-m", "armv7e-m"]:
        return "arm32"

    if arch in ["arm64", "aarch64", "armv8", "armv8-a", "armv8.2-a", "armv8.3-a", "armv8-m", "armv8-r", "armv8.1-a", "armv8.4-a", "armv8.5-a", "armv8.6-a", "armv8-m.base", "armv8-m.main", "armv8.1-m.main", "armv9", "armv9-a", "iwmmxt", "iwmmxt2"]:
        return "arm64"

    return ""



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import typing as T

from .utils import makedirs
from .config import get_global_config
from .linkers import get_all_linker_types
from .archivers import get_all_archiver_types
from .shared_linkers import get_all_shared_linker_types
from .compilers import get_all_c_compiler_types, get_all_cpp_compiler_types, get_all_as_compiler_types, get_all_asm_compiler_types


def dp(string: T.Union[str, None]) -> str:
    if string is None:
        return "Let the program decide"
    return string


def add_tool_dict(dictionary: T.Dict[T.Any, T.Any], tool: T.List[T.Union[str, None]], tool_name: str) -> None:
    if tool[0] is not None:
        if tool_name not in dictionary:
            dictionary[tool_name] = {}
        dictionary[tool_name]["type"] = tool[0]
    if tool[1] is not None:
        if tool_name not in dictionary:
            dictionary[tool_name] = {}
        dictionary[tool_name]["path"] = tool[1]


def multiple_choices(question: str, choices: T.List[T.Union[str, None]], values: T.Union[T.List[T.Any], None] = None) -> T.Any:
    if values is None:
        values = choices
    assert len(choices) == len(values)

    answer = 0
    while answer not in range(1, len(choices) + 1):
        print("\033[H\033[2J", end="")
        print(question)
        for i in range(1, len(choices) + 1):
            print(f"[{i}]: {dp(choices[i - 1])}")
        answer_str = input(f"{' '.join([str(i) for i in range(1, len(choices) + 1)])}: ")
        if answer_str.isnumeric():
            answer = int(answer_str)

    return values[answer - 1]


class InteractiveConfig:
    def __init__(self, global_config: T.Union[str, None] = None, local_config: str = "./powermake_config"):
        self.target_architecture: T.Union[str, None] = None
        self.target_operating_system: T.Union[str, None] = None
        self.c_compiler: T.List[T.Union[str, None]] = [None, None]
        self.cpp_compiler: T.List[T.Union[str, None]] = [None, None]
        self.as_compiler: T.List[T.Union[str, None]] = [None, None]
        self.asm_compiler: T.List[T.Union[str, None]] = [None, None]
        self.archiver: T.List[T.Union[str, None]] = [None, None]
        self.linker: T.List[T.Union[str, None]] = [None, None]
        self.shared_linker: T.List[T.Union[str, None]] = [None, None]

        self.global_config = global_config
        self.local_config = local_config
        answer = 0
        while answer != 4:
            choices: T.List[T.Union[str, None]] = [
                f"Target operating system ({dp(self.target_operating_system)})",
                f"Target architecture ({dp(self.target_architecture)})",
                "Toolchain\n",
                "Save configuration"
            ]
            answer = multiple_choices("What do you want to configure ?", choices, [i for i in range(1, len(choices) + 1)])

            if answer == 1:
                self.target_operating_system = multiple_choices("Select the target operating system", [None, "Linux", "Windows", "MacOS", "Write my own string"])
                if self.target_operating_system == "Write my own string":
                    self.target_operating_system = input("Operating System: ")
            elif answer == 2:
                self.target_architecture = multiple_choices("Select the target architecture", [None, "x86", "x64", "arm32", "arm64"])
            elif answer == 3:
                self.toolchain_menu()

        answer = multiple_choices("In which configuration do you want to write this ?", ["Local config", "Global config"], [1, 2])
        if answer == 1:
            self.save_config(local_config)
        else:
            self.save_config(global_config or get_global_config())

    def toolchain_menu(self) -> None:
        answer = 0
        choices: T.List[T.Union[str, None]] = [
            "C compiler",
            "C++ compiler",
            "AS compiler (.s and .S files)",
            "ASM compiler (.asm files)",
            "Archiver (to make static libraries)",
            "Shared Linker (to make shared libraries)",
            "Linker\n",
            "Back to main menu"
        ]
        while answer != len(choices):
            answer = multiple_choices("What do you want to configure ?", choices, [i for i in range(1, len(choices) + 1)])

            if answer == 1:
                self.tool_menu(self.c_compiler, "C compiler", get_all_c_compiler_types())
            elif answer == 2:
                self.tool_menu(self.cpp_compiler, "C++ compiler", get_all_cpp_compiler_types())
            elif answer == 3:
                self.tool_menu(self.as_compiler, "AS compiler", get_all_as_compiler_types())
            elif answer == 4:
                self.tool_menu(self.asm_compiler, "ASM compiler", get_all_asm_compiler_types())
            elif answer == 5:
                self.tool_menu(self.archiver, "Archiver", get_all_archiver_types())
            elif answer == 6:
                self.tool_menu(self.shared_linker, "Shared Linker", get_all_shared_linker_types())
            elif answer == 7:
                self.tool_menu(self.linker, "Linker", get_all_linker_types())

    def tool_menu(self, tool: T.List[T.Union[str, None]], tool_name: str, tool_types: T.Set[str]) -> None:
        answer = 0
        while answer != 3:
            choices: T.List[T.Union[str, None]] = [
                f"{tool_name} type (what syntax) ({dp(tool[0])})",
                f"{tool_name} path ({dp(tool[1])})\n",
                "Back to toolchain menu"
            ]
            answer = multiple_choices("What do you want to configure ?", choices, [i for i in range(1, len(choices) + 1)])

            if answer == 1:
                tool[0] = multiple_choices(f"Select the {tool_name} type", [None] + list(tool_types))
            elif answer == 2:
                print(f"Enter the path of the {tool_name} (Empty to let the program decide)")
                tool[1] = input("Path: ")
                if tool[1] == "":
                    tool[1] = None

    def save_config(self, filepath: str) -> None:
        config = {}
        if os.path.isfile(filepath):
            verif = 'n'
            while verif == 'n':
                answer = multiple_choices(f"Their is already a file at {filepath}\nWhat should we do ?", ["Merge", "Overwrite"], [1, 2])
                if answer == 2:
                    verif = ''
                    while verif != 'y' and verif != 'n':
                        verif = input("Are you sure you want to erase all your precedent configuration ?\ny/n: ").lower()
                else:
                    verif = 'y'
            if answer == 1:
                with open(filepath, "r") as file:
                    config = json.load(file)

        if self.target_operating_system is not None:
            config["target_operating_system"] = self.target_operating_system

        if self.target_architecture is not None:
            config["target_architecture"] = self.target_architecture

        add_tool_dict(config, self.c_compiler, "c_compiler")
        add_tool_dict(config, self.cpp_compiler, "cpp_compiler")
        add_tool_dict(config, self.as_compiler, "as_compiler")
        add_tool_dict(config, self.asm_compiler, "asm_compiler")
        add_tool_dict(config, self.archiver, "archiver")
        add_tool_dict(config, self.shared_linker, "shared_linker")
        add_tool_dict(config, self.linker, "linker")

        makedirs(os.path.dirname(filepath), exist_ok=True)
        with open(filepath, "w") as file:
            json.dump(config, file, indent=4)

        print("Configuration saved at", filepath)



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import typing as T

from ..utils import get_empty_file
from .common import SharedLinker


class SharedLinkerGNU(SharedLinker):
    type: T.ClassVar = "gnu"
    shared_lib_extension: T.ClassVar = ".so"

    def __init__(self, path: str = "c++"):
        super().__init__(path)

    def format_args(self, shared_libs: T.List[str], flags: T.List[str]) -> T.List[str]:
        return ["-l" + lib for lib in shared_libs] + self.translate_flags(flags)

    def basic_link_command(self, outputfile: str, objectfiles: T.Iterable[str], archives: T.List[str] = [], args: T.List[str] = []) -> T.List[str]:
        return [self.path, "-shared", "-o", outputfile, *objectfiles, *archives, *args]
    
    def check_if_arg_exists(self, arg: str) -> bool:
        return subprocess.run([self.path, arg, "-E", get_empty_file()], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL).returncode == 0


class SharedLinkerLD(SharedLinkerGNU):
    type: T.ClassVar = "ld"

    def __init__(self, path: str = "ld"):
        super().__init__(path)

    def check_if_arg_exists(self, arg: str) -> bool:
        return subprocess.run([self.path, arg, "-w"], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL).returncode == 0


class SharedLinkerGCC(SharedLinkerGNU):
    type: T.ClassVar = "gcc"

    def __init__(self, path: str = "gcc"):
        super().__init__(path)


class SharedLinkerGPlusPlus(SharedLinkerGNU):
    type: T.ClassVar = "g++"

    def __init__(self, path: str = "g++"):
        super().__init__(path)


class SharedLinkerClang(SharedLinkerGNU):
    type: T.ClassVar = "clang"

    def __init__(self, path: str = "clang"):
        super().__init__(path)


class SharedLinkerClangPlusPlus(SharedLinkerGNU):
    type: T.ClassVar = "clang++"

    def __init__(self, path: str = "clang++"):
        super().__init__(path)



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import json
import ctypes
import string
import platform
import tempfile
import subprocess
import typing as T

from .utils import makedirs


def get_drives() -> T.List[str]:
    if platform.platform().lower().startswith("win") and hasattr(ctypes, 'windll'):
        drives = []
        bitmask = ctypes.windll.kernel32.GetLogicalDrives()
        for letter in string.ascii_uppercase:
            if bitmask & 1:
                drives.append(letter + ":\\")
            bitmask >>= 1

        return drives

    return []


vcvars = [
    "path",
    "lib",
    "libpath",
    "include",
    "DevEnvdir",
    "VSInstallDir",
    "VCInstallDir",
    "WindowsSdkDir",
    "WindowsLibPath",
    "WindowsSDKVersion",
    "WindowsSdkBinPath",
    "WindowsSdkVerBinPath",
    "ExtensionSdkDir",
    "UniversalCRTSdkDir",
    "UCRTVersion",
    "VCToolsVersion",
    "VCIDEInstallDir",
    "VCToolsInstallDir",
    "VCToolsRedistDir",
    "VisualStudioVersion",
    "VSCMD_VER",
    "VSCMD_ARG_app_plat",
    "VSCMD_ARG_HOST_ARCH",
    "VSCMD_ARG_TGT_ARCH"
]


vsvers = {
    "17.0": "2022",
    "16.0": "2019",
    "15.0": "2017",
    "14.0": "2015",
    "12.0": "2013",
    "11.0": "2012",
    "10.0": "2010",
    "9.0": "2008",
    "8.0": "2005",
    "7.1": "2003",
    "7.0": "7.0",
    "6.0": "6.0",
    "5.0": "5.0",
    "4.2": "4.2"
}


def _load_vcvarsall(vcvarsall_path: str, version: str, architecture: str) -> T.Dict[str, str]:
    tempdir = tempfile.TemporaryDirectory("powermake")
    genvcvars_filepath = os.path.join(tempdir.name, "genvcvars.bat")
    file = open(genvcvars_filepath, "w")

    file.write("@echo off\n")
    file.write("chcp 65001 > nul\n")

    if float(version) >= 16:
        file.write("set VSCMD_SKIP_SENDTELEMETRY=yes\n")

    file.write("call \"%s\" %s > nul\n" % (vcvarsall_path, architecture))

    for var in vcvars:
        file.write("echo " + var + " = %" + var + "%\n")

    file.close()

    lines = subprocess.check_output([genvcvars_filepath], encoding="ascii").split("\n")

    tempdir.cleanup()

    variables = {}
    for line in lines:
        if "=" not in line:
            continue
        key, value = line.split("=")
        key = key.strip()
        value = value.strip()

        variables[key] = value

    return variables


def _find_vcvarsall() -> T.Tuple[T.Union[str, None], T.Union[str, None]]:
    for version in vsvers:
        for logical_drive in get_drives():
            paths = []

            for prg_dir in ("Program Files (x86)", "Program Files"):
                temp_path = os.path.join(logical_drive, prg_dir, "Microsoft Visual Studio", vsvers[version])
                if not os.path.isdir(temp_path):
                    continue
                for dirname in os.listdir(temp_path):
                    if os.path.isdir(os.path.join(temp_path, dirname)):
                        paths.append(os.path.join(temp_path, dirname, "VC", "Auxiliary", "Build"))
                paths.append(os.path.join(logical_drive, prg_dir, "Microsoft Visual Studio " + version, "VC"))

            if version == "6.0":
                paths.append(os.path.join(logical_drive, "Program Files", "Microsoft Visual Studio", "VC98", "Bin"))
                paths.append(os.path.join(logical_drive, "Program Files (x86)", "Microsoft Visual Studio", "VC98", "Bin"))

            for path in paths:
                filepath = os.path.join(path, "vcvarsall.bat")
                if os.path.exists(filepath):
                    return filepath, version
                filepath = os.path.join(path, "vcvars32.bat")
                if os.path.exists(filepath):
                    return filepath, version

    return None, None


def is_msvc_loaded_correctly(env: T.Dict[str, str]) -> bool:
    for var in vcvars:
        if var not in env:
            return False
    return True


def load_envs_from_file(filepath: str, architecture: str = "x86") -> T.Dict[str, T.Any]:
    try:
        with open(filepath, "r") as file:
            return dict(json.load(file))
    except OSError:
        return {}


def store_envs_to_file(filepath: str, envs: T.Dict[str, T.Any]) -> None:
    makedirs(os.path.dirname(filepath), exist_ok=True)
    with open(filepath, "w") as file:
        json.dump(envs, file, indent=4)


def load_msvc_environment(storage_path: str, architecture: str = "x86") -> T.Union[T.Dict[str, str], None]:
    envs = load_envs_from_file(storage_path)
    if architecture in envs and is_msvc_loaded_correctly(envs[architecture]):
        return dict(envs[architecture])

    if "vcvarsall_path" not in envs or "vcversion" not in envs:
        vcvarsall_path, vcversion = _find_vcvarsall()
        if vcvarsall_path is None:
            return None
        envs["vcvarsall_path"] = vcvarsall_path
        envs["vcversion"] = vcversion

    env = _load_vcvarsall(envs["vcvarsall_path"], envs["vcversion"], architecture)
    if env is not None:
        envs[architecture] = env

    store_envs_to_file(storage_path, envs)

    return env



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
import shutil
import typing as T


_powermake_flags_to_gnu_flags: T.Dict[str, T.List[str]] = {
    "-Wsecurity": ["-Wall", "-Wextra", "-fanalyzer", "-Wformat-security", "-Wformat", "-Wconversion", "-Wsign-conversion", "-Wtrampolines", "-Wbidi-chars=any,ucn"],
    "-fsecurity=1": ["-Wsecurity", "-fstrict-flex-arrays=2", "-fcf-protection=full", "-mbranch-protection=standard", "-Wl,-z,nodlopen", "-Wl,-z,noexecstack", "-Wl,-z,relro", "-fPIE", "-pie", "-fPIC", "-fno-delete-null-pointer-checks", "-fno-strict-overflow", "-fno-strict-aliasing", "-fexceptions", "-Wl,--as-needed", "-Wl,--no-copy-dt-needed-entries"],
    "-fsecurity=2": ["-fsecurity=1", "-D_FORTIFY_SOURCE=3", "-D_GLIBCXX_ASSERTIONS", "-fstack-clash-protection", "-fstack-protector-strong", "-Wl,-z,now", "-ftrivial-auto-var-init=zero"],
    "-fsecurity": ["-fsecurity=2"],
    "-Weverything": ["-Weverything", "-Wsecurity", "-pedantic", "-Wsuggest-attribute=pure", "-Wsuggest-attribute=const", "-Wsuggest-attribute=noreturn", "-Wsuggest-attribute=malloc", "-Wsuggest-attribute=returns_nonnull", "-Wsuggest-attribute=format", "-Wmissing-format-attribute", "-Wsuggest-attribute=cold", "-Waggregate-return", "-Wduplicated-branches", "-Wduplicated-cond", "-Wflex-array-member-not-at-end", "-Wfloat-equal", "-Wformat-nonliteral", "-Wformat-signedness", "-Wformat-y2k", "-Winit-self", "-Winvalid-utf8", "-Wjump-misses-init", "-Wlogical-op", "-Wmissing-declarations", "-Wmissing-include-dirs", "-Wmissing-prototypes", "-Wmissing-variable-declarations", "-Wmultichar", "-Wnested-externs", "-Wnull-dereference", "-Wopenacc-parallelism", "-Wredundant-decls", "-Wshadow", "-Wstack-protector", "-Wstrict-flex-arrays=3", "-Wstrict-prototypes", "-Wsuggest-final-methods", "-Wsuggest-final-types", "-Wswitch-default", "-Wundef", "-Wunsuffixed-float-constants", "-Wunused-macros", "-Wuseless-cast", "-Wvector-operation-performance", "-Wwrite-strings"],
    "-ffuzzer": ["-fsanitize=address,fuzzer"],
}


class Tool(abc.ABC):
    type: T.ClassVar = ""
    default_path: T.ClassVar = ""
    translation_dict: T.ClassVar[T.Dict[str, T.List[str]]] = _powermake_flags_to_gnu_flags

    def __init__(self, path: str) -> None:
        self._name = path
        self.verified_translation_dict: T.Dict[str, T.List[str]] = {}
        self.reload()

    def is_available(self) -> bool:
        return self.path != ""

    def reload(self, new_path: T.Union[str, None] = None) -> None:
        if new_path is not None:
            self._name = new_path
        path = shutil.which(self._name)
        if path is None:
            self.path = ""
        else:
            self.path = path

    @abc.abstractmethod
    def check_if_arg_exists(self, arg: str) -> bool:
        return False
    
    def _translate_flag(self, flag: str, output_list: T.List[str], already_translated_flags: T.List[str]) -> None:
        already_translated_flags.append(flag)

        if flag in self.verified_translation_dict:
            output_list.extend([f for f in self.verified_translation_dict[flag] if f not in output_list])
            return

        if flag not in self.translation_dict:
            # This flag is not in any translation table, we left it untouched.
            # We don't check if flag is in output_list, that should have been done before and if it's not the case, it's probably that the user enforced that
            output_list.append(flag)
            return

        self.verified_translation_dict[flag] = []

        for f in self.translation_dict[flag]:
            if f not in already_translated_flags and f in self.translation_dict:
                self._translate_flag(f, output_list, already_translated_flags)
                # after this call, self.verified_translation_dict will contain an entry for the key f, containing all flags flattened.
                self.verified_translation_dict[flag].extend(self.verified_translation_dict[f])
            else:
                if self.check_if_arg_exists(f):
                    # the flag f is valid and is not a combination of flags
                    self.verified_translation_dict[flag].append(f)
                    if f not in output_list:
                        output_list.append(f)

    def translate_flags(self, flags: T.List[str]) -> T.List[str]:
        translated_flags: T.List[str] = []
        already_translated_flags: T.List[str] = []
        for flag in flags:
            self._translate_flag(flag, translated_flags, already_translated_flags)

        return translated_flags



def load_tool_tuple_from_file(conf: T.Dict[str, T.Any], tool_name: str, object_getter: T.Callable[[str], T.Union[T.Callable[[], Tool], None]], tool_list_getter: T.Callable[[], T.Set[str]]) -> T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Tool]], None]:
    if tool_name not in conf:
        return None

    if "path" in conf[tool_name] or "type" in conf[tool_name]:
        if "type" in conf[tool_name]:
            tool_type = conf[tool_name]["type"].lower()
        else:
            tool_type = "gnu"

        ObjectConstructor = object_getter(tool_type)
        if ObjectConstructor is None:
            raise ValueError("Unsupported %s type: %s\n\nShould be one of them: %s" % (tool_name, tool_type, " ".join(tool_list_getter())))

        if "path" in conf[tool_name]:
            tool_path = conf[tool_name]["path"]
        else:
            tool_path = None

        return (tool_path, ObjectConstructor)

    return None


def load_tool_from_tuple(tool_tuple: T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Tool]], None], tool_name: str) -> T.Union[Tool, None]:
    if tool_tuple is not None:
        tool: Tool
        if tool_tuple[0] is None:
            tool = tool_tuple[1]()
        else:
            tool = T.cast(T.Callable[[str], Tool], tool_tuple[1])(tool_tuple[0])

        if not tool.is_available():
            if tool_tuple[0] is None:
                tool_path = tool.type
            else:
                tool_path = tool_tuple[0]
            raise ValueError("The %s %s could not be found on your machine" % (tool_name, tool_path))

        return tool
    return None


def find_tool(object_getter: T.Callable[[str], T.Union[T.Callable[[], Tool], None]], *tool_types: str) -> T.Union[Tool, None]:
    for tool_type in tool_types:
        ObjectConstructor = object_getter(tool_type)
        if ObjectConstructor is None:
            continue
        tool: Tool = ObjectConstructor()
        if tool is not None and tool.is_available():
            return tool
    return None



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "Archiver", "ArchiverGNU", "ArchiverAR", "ArchiverLLVM_AR", "ArchiverMSVC"
]

import typing as T

from .common import Archiver
from .gnu import ArchiverGNU, ArchiverAR, ArchiverLLVM_AR
from .msvc import ArchiverMSVC


_archiver_types: T.Dict[str, T.Callable[[], Archiver]] = {
    "gnu": ArchiverGNU,
    "ar": ArchiverAR,
    "llvm-ar": ArchiverLLVM_AR,
    "msvc": ArchiverMSVC
}


def GenericArchiver(archiver_type: str) -> T.Union[T.Callable[[], Archiver], None]:
    if archiver_type not in _archiver_types:
        return None
    return _archiver_types[archiver_type]


def get_all_archiver_types() -> T.Set[str]:
    return set(_archiver_types.keys())



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import colorama
import typing as T


def init_colors() -> None:
    colorama.init()


def print_info(string: T.Any, verbosity: int, counter: int = 0, total: int = 0) -> None:
    if verbosity >= 1:
        if counter == 0:
            print(colorama.Fore.LIGHTMAGENTA_EX + str(string) + colorama.Style.RESET_ALL)
        elif total != 0:
            print(colorama.Fore.LIGHTMAGENTA_EX + f"[{counter}/{total}] {round(100*counter/total)}% " + str(string) + colorama.Style.RESET_ALL)
        else:
            print(colorama.Fore.LIGHTMAGENTA_EX + f"[{counter}/-] " + str(string) + colorama.Style.RESET_ALL)


def print_debug_info(string: T.Any, verbosity: int) -> None:
    if verbosity >= 2:
        print(colorama.Fore.LIGHTBLACK_EX + str(string) + colorama.Style.RESET_ALL)



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import copy
import json
import argparse
import platform
import typing as T

from .display import print_debug_info
from .tools import load_tool_tuple_from_file, load_tool_from_tuple, find_tool
from .search_visual_studio import load_msvc_environment
from .architecture import simplify_architecture
from .compilers import Compiler, CompilerGNU, CompilerGNUPlusPlus, GenericCompiler, get_all_c_compiler_types, get_all_cpp_compiler_types, get_all_as_compiler_types, get_all_asm_compiler_types
from .archivers import Archiver, GenericArchiver, get_all_archiver_types
from .linkers import Linker, LinkerGNU, GenericLinker, get_all_linker_types
from .shared_linkers import SharedLinker, GenericSharedLinker, get_all_shared_linker_types


def get_global_config() -> str:
    global_config = os.getenv("POWERMAKE_CONFIG")
    if global_config is None:
        global_config = os.path.normpath(os.path.expanduser("~/.powermake/powermake_config.json"))
    return global_config


def auto_toolchain(preference: T.Union[str, None], c_compiler: T.Union[Compiler, None], cpp_compiler: T.Union[Compiler, None], as_compiler: T.Union[Compiler, None], asm_compiler: T.Union[Compiler, None], archiver: T.Union[Archiver, None], linker: T.Union[Linker, None], shared_linker: T.Union[SharedLinker, None]) -> T.Tuple[str, str, str, str, str, str, str]:
    all_None = True
    if c_compiler is not None:
        c_compiler_type = c_compiler.type
        all_None = False
    else:
        c_compiler_type = None

    if cpp_compiler is not None:
        cpp_compiler_type = cpp_compiler.type
        all_None = False
    else:
        cpp_compiler_type = None

    if as_compiler is not None:
        as_compiler_type = as_compiler.type
        all_None = False
    else:
        as_compiler_type = None

    if asm_compiler is not None:
        asm_compiler_type = asm_compiler.type
        all_None = False
    else:
        asm_compiler_type = None

    if archiver is not None:
        archiver_type = archiver.type
        all_None = False
    else:
        archiver_type = None

    if linker is not None:
        linker_type = linker.type
        all_None = False
    else:
        linker_type = None

    if shared_linker is not None:
        shared_linker_type = shared_linker.type
        all_None = False
    else:
        shared_linker_type = None

    if all_None and preference == "gnu" or c_compiler_type == "gnu" or cpp_compiler_type == "gnu++" or as_compiler_type == "gnu" or linker_type in ("gnu", "gnu++") or shared_linker_type in ("gnu", "gnu++") or archiver_type == "gnu":
        # GNU toolchain
        c_compiler_type = "gnu"
        cpp_compiler_type = "gnu++"
        as_compiler_type = "gnu"
        archiver_type = "gnu"
        linker_type = "gnu++"
        shared_linker_type = "gnu++"
    elif all_None and preference == "gcc" or c_compiler_type == "gcc" or cpp_compiler_type == "g++" or as_compiler_type == "gcc" or linker_type in ("gcc", "g++") or shared_linker_type in ("gcc", "g++") or archiver_type == "ar":
        # GCC toolchain
        c_compiler_type = "gcc"
        cpp_compiler_type = "g++"
        as_compiler_type = "gcc"
        archiver_type = "ar"
        linker_type = "g++"
        shared_linker_type = "g++"
    elif all_None and preference == "clang" or c_compiler_type == "clang" or cpp_compiler_type == "clang++" or as_compiler_type == "clang" or linker_type in ("clang", "clang++") or shared_linker_type in ("clang", "clang++") or archiver_type == "llvm-ar":
        # clang toolchain
        c_compiler_type = "clang"
        cpp_compiler_type = "clang++"
        as_compiler_type = "clang"
        archiver_type = "llvm-ar"
        linker_type = "clang++"
        shared_linker_type = "clang++"
    elif all_None and preference == "msvc" or c_compiler_type == "msvc" or cpp_compiler_type == "msvc" or linker_type == "msvc" or shared_linker_type == "msvc" or archiver_type == "msvc":
        # MSVC toolchain
        c_compiler_type = "msvc"
        cpp_compiler_type = "msvc"
        as_compiler_type = "clang"
        archiver_type = "msvc"
        linker_type = "msvc"
        shared_linker_type = "msvc"

    return (c_compiler_type, cpp_compiler_type, as_compiler_type, asm_compiler_type, archiver_type, linker_type, shared_linker_type)


def get_toolchain_tuple(path: T.Union[str, None, bool]) -> T.Union[T.Tuple[str, str], None]:
    if not path or path is True:
        return None
    if path.endswith("gcc"):
        return ("gcc", path[:-3])
    if path.endswith("clang"):
        return ("clang", path[:-5])
    if path.endswith("clang++"):
        return ("clang", path[:-7])
    if path.endswith("g++"):
        return ("gcc", path[:-3])
    if path.endswith("ar"):
        return ("gnu", path[:-2])
    if path.endswith("ld"):
        return ("gnu", path[:-2])
    if path.endswith("cc"):
        return ("gnu", path[:-2])
    if path.endswith("cpp"):
        return ("gnu", path[-3])
    if path.endswith("c++"):
        return ("gnu", path[-3])
    
    return None


def replace_architecture(string: str, new_arch: str) -> str:
    new_arch = "/" + new_arch + "/"
    for arch in ("/x86/", "/x64/", "/arm32/", "/arm64/"):
        string = string.replace(arch, new_arch)
    return string


class Config:
    def __init__(self, target_name: str, *, args_parsed: T.Union[argparse.Namespace, None] = None, debug: bool = False, rebuild: bool = False, verbosity: int = 1, nb_jobs: int = 0, single_file: T.Union[str, None] = None, compile_commands_dir: T.Union[str, None] = None, local_config: T.Union[str, None] = "./powermake_config.json", global_config: T.Union[str, None] = None) -> None:
        """Create an object that loads all configurations files and search for compilers.

        This objects hold all the configuration for the compilation.

        If you want to compile multiple set of files with different settings, you have to make a copy of this object and modify the copy.
        """
        self.target_name = target_name
        self._args_parsed = args_parsed
        self.verbosity = verbosity
        self.debug = debug
        self.rebuild = rebuild
        self.nb_jobs = nb_jobs
        self.single_file = single_file
        self.compile_commands_dir = compile_commands_dir
        self.nb_total_operations = 0

        self.c_compiler: T.Union[Compiler, None] = None
        self.cpp_compiler: T.Union[Compiler, None] = None
        self.as_compiler: T.Union[Compiler, None] = None
        self.asm_compiler: T.Union[Compiler, None] = None
        self.archiver: T.Union[Archiver, None] = None
        self.linker: T.Union[Linker, None] = None
        self.shared_linker: T.Union[SharedLinker, None] = None

        self.target_operating_system: str = ""
        self.host_operating_system: str = ""

        self.target_architecture: str = ""
        self.target_simplified_architecture: str = ""
        self.host_architecture: str = ""
        self.host_simplified_architecture: str = ""

        self.obj_build_directory: str = ""
        self.exe_build_directory: str = ""
        self.lib_build_directory: str = ""

        self.defines: T.List[str] = []
        self.shared_libs: T.List[str] = []
        self.additional_includedirs: T.List[str] = []
        self.c_flags: T.List[str] = []
        self.cpp_flags: T.List[str] = []
        self.as_flags: T.List[str] = []
        self.asm_flags: T.List[str] = []
        self.ar_flags: T.List[str] = []
        self.ld_flags: T.List[str] = []
        self.shared_linker_flags: T.List[str] = []

        self.exported_headers: T.List[T.Tuple[str, T.Union[str, None]]] = []

        c_compiler_tuple: T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Compiler]], None] = None
        cpp_compiler_tuple: T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Compiler]], None] = None
        as_compiler_tuple: T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Compiler]], None] = None
        asm_compiler_tuple: T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Compiler]], None] = None
        archiver_tuple: T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Archiver]], None] = None
        linker_tuple: T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Linker]], None] = None
        shared_linker_tuple: T.Union[T.Tuple[T.Union[str, None], T.Callable[[], SharedLinker]], None] = None

        if global_config is None:
            global_config = get_global_config()

        self.global_config_dir = os.path.dirname(global_config)

        for path in (local_config, global_config):
            if path is None:
                continue
            try:
                with open(path, "r") as file:
                    conf: T.Dict[str, T.Any] = json.load(file)
                    if c_compiler_tuple is None:
                        c_compiler_tuple = T.cast(T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Compiler]], None], load_tool_tuple_from_file(conf, "c_compiler", GenericCompiler, get_all_c_compiler_types))

                    if cpp_compiler_tuple is None:
                        cpp_compiler_tuple = T.cast(T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Compiler]], None], load_tool_tuple_from_file(conf, "cpp_compiler", GenericCompiler, get_all_cpp_compiler_types))

                    if as_compiler_tuple is None:
                        as_compiler_tuple = T.cast(T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Compiler]], None], load_tool_tuple_from_file(conf, "as_compiler", GenericCompiler, get_all_as_compiler_types))

                    if asm_compiler_tuple is None:
                        asm_compiler_tuple = T.cast(T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Compiler]], None], load_tool_tuple_from_file(conf, "asm_compiler", GenericCompiler, get_all_asm_compiler_types))

                    if archiver_tuple is None:
                        archiver_tuple = T.cast(T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Archiver]], None], load_tool_tuple_from_file(conf, "archiver", GenericArchiver, get_all_archiver_types))

                    if linker_tuple is None:
                        linker_tuple = T.cast(T.Union[T.Tuple[T.Union[str, None], T.Callable[[], Linker]], None], load_tool_tuple_from_file(conf, "linker", GenericLinker, get_all_linker_types))

                    if shared_linker_tuple is None:
                        shared_linker_tuple = T.cast(T.Union[T.Tuple[T.Union[str, None], T.Callable[[], SharedLinker]], None], load_tool_tuple_from_file(conf, "shared_linker", GenericSharedLinker, get_all_shared_linker_types))

                    if self.target_operating_system == "" and "target_operating_system" in conf:
                        self.target_operating_system = conf["target_operating_system"]
                    if self.host_operating_system == "" and "host_operating_system" in conf:
                        self.host_operating_system = conf["host_operating_system"]

                    if self.target_architecture == "" and "target_architecture" in conf:
                        self.target_architecture = conf["target_architecture"]
                    if self.host_architecture == "" and "host_architecture" in conf:
                        self.host_architecture = conf["host_architecture"]

                    if self.obj_build_directory == "" and "obj_build_directory" in conf:
                        self.obj_build_directory = conf["obj_build_directory"]
                    if self.exe_build_directory == "" and "exe_build_directory" in conf:
                        self.exe_build_directory = conf["exe_build_directory"]
                    if self.lib_build_directory == "" and "lib_build_directory" in conf:
                        self.lib_build_directory = conf["lib_build_directory"]

                    if self.nb_jobs == 0 and "nb_jobs" in conf:
                        self.nb_jobs = conf["nb_jobs"]

                    if self.compile_commands_dir is None and "compile_commands_dir" in conf:
                        self.compile_commands_dir = conf["compile_commands_dir"]

                    if "defines" in conf and isinstance(conf["defines"], list):
                        for define in conf["defines"]:
                            if isinstance(define, str) and define not in self.defines:
                                self.defines.append(define)

                    if "shared_libs" in conf and isinstance(conf["shared_libs"], list):
                        for shared_lib in conf["shared_libs"]:
                            if isinstance(shared_lib, str) and shared_lib not in self.shared_libs:
                                self.shared_libs.append(shared_lib)

                    if "additional_includedirs" in conf and isinstance(conf["additional_includedirs"], list):
                        for includedir in conf["additional_includedirs"]:
                            if isinstance(includedir, str) and includedir not in self.additional_includedirs:
                                self.additional_includedirs.append(includedir)

                    if "c_flags" in conf and isinstance(conf["c_flags"], list):
                        for c_flag in conf["c_flags"]:
                            if isinstance(c_flag, str) and c_flag not in self.c_flags:
                                self.c_flags.append(c_flag)

                    if "cpp_flags" in conf and isinstance(conf["cpp_flags"], list):
                        for cpp_flag in conf["cpp_flags"]:
                            if isinstance(cpp_flag, str) and cpp_flag not in self.cpp_flags:
                                self.cpp_flags.append(cpp_flag)

                    if "c_cpp_flags" in conf and isinstance(conf["c_cpp_flags"], list):
                        for c_cpp_flag in conf["c_cpp_flags"]:
                            if isinstance(c_cpp_flag, str):
                                if c_cpp_flag not in self.c_flags:
                                    self.c_flags.append(c_cpp_flag)
                                if c_cpp_flag not in self.cpp_flags:
                                    self.cpp_flags.append(c_cpp_flag)

                    if "c_cpp_as_asm_flags" in conf and isinstance(conf["c_cpp_as_asm_flags"], list):
                        for c_cpp_as_asm_flag in conf["c_cpp_as_asm_flags"]:
                            if isinstance(c_cpp_as_asm_flag, str):
                                if c_cpp_as_asm_flag not in self.c_flags:
                                    self.c_flags.append(c_cpp_as_asm_flag)
                                if c_cpp_as_asm_flag not in self.cpp_flags:
                                    self.cpp_flags.append(c_cpp_as_asm_flag)
                                if c_cpp_as_asm_flag not in self.as_flags:
                                    self.as_flags.append(c_cpp_as_asm_flag)
                                if c_cpp_as_asm_flag not in self.asm_flags:
                                    self.asm_flags.append(c_cpp_as_asm_flag)

                    if "as_flags" in conf and isinstance(conf["as_flags"], list):
                        for as_flag in conf["as_flags"]:
                            if isinstance(as_flag, str) and as_flag not in self.as_flags:
                                self.as_flags.append(as_flag)

                    if "asm_flags" in conf and isinstance(conf["asm_flags"], list):
                        for asm_flag in conf["asm_flags"]:
                            if isinstance(asm_flag, str) and asm_flag not in self.asm_flags:
                                self.asm_flags.append(asm_flag)

                    if "ar_flags" in conf and isinstance(conf["ar_flags"], list):
                        for ar_flag in conf["ar_flags"]:
                            if isinstance(ar_flag, str) and ar_flag not in self.ar_flags:
                                self.ar_flags.append(ar_flag)

                    if "ld_flags" in conf and isinstance(conf["ld_flags"], list):
                        for ld_flag in conf["ld_flags"]:
                            if isinstance(ld_flag, str) and ld_flag not in self.ld_flags:
                                self.ld_flags.append(ld_flag)

                    if "shared_linker_flags" in conf and isinstance(conf["shared_linker_flags"], list):
                        for shared_linker_flag in conf["shared_linker_flags"]:
                            if isinstance(shared_linker_flag, str) and shared_linker_flag not in self.shared_linker_flags:
                                self.shared_linker_flags.append(shared_linker_flag)

                    if "exported_headers" in conf and isinstance(conf["exported_headers"], list):
                        for exported_header in conf["exported_headers"]:
                            if isinstance(exported_header, str) and (exported_header, None) not in self.exported_headers:
                                self.exported_headers.append((exported_header, None))
                            elif isinstance(exported_header, list) and len(exported_header) == 2 and tuple(exported_header) not in self.exported_headers:
                                self.exported_headers.append((exported_header[0], exported_header[1]))

            except OSError:
                pass

        if self.nb_jobs == 0:
            self.nb_jobs = os.cpu_count() or 8
        
        target_os_autodetected = False
        if self.target_operating_system == "":
            target_os_autodetected = True
            self.target_operating_system = platform.system()
        if self.host_operating_system == "":
            self.host_operating_system = platform.system()

        if self.target_architecture == "":
            self.target_architecture = platform.machine()
        if self.host_architecture == "":
            self.host_architecture = platform.machine()

        self.set_target_architecture(self.target_architecture, reload_tools_and_build_dir=False)

        self.c_compiler = T.cast(T.Union[Compiler, None], load_tool_from_tuple(c_compiler_tuple, "compiler"))
        self.cpp_compiler = T.cast(T.Union[Compiler, None], load_tool_from_tuple(cpp_compiler_tuple, "compiler"))
        self.as_compiler = T.cast(T.Union[Compiler, None], load_tool_from_tuple(as_compiler_tuple, "compiler"))
        self.asm_compiler = T.cast(T.Union[Compiler, None], load_tool_from_tuple(asm_compiler_tuple, "compiler"))
        self.archiver = T.cast(T.Union[Archiver, None], load_tool_from_tuple(archiver_tuple, "archiver"))
        self.linker = T.cast(T.Union[Linker, None], load_tool_from_tuple(linker_tuple, "linker"))
        self.shared_linker = T.cast(T.Union[SharedLinker, None], load_tool_from_tuple(shared_linker_tuple, "shared_linker"))

        preference = None
        if "CCC_ANALYZER_ANALYSIS" in os.environ:
            # This is the signature of scan-build
            # if no toolchain is specified, the auto toolchain should prefer clang
            preference = "clang"
        else:
            compiler_path = os.getenv("CC") or os.getenv("CXX")
            if compiler_path is not None:
                if "clang" in compiler_path:
                    preference = "clang"
                elif "gcc" in compiler_path or "g++" in compiler_path:
                    preference = "gcc"
                elif "cl" in compiler_path:
                    preference = "msvc"
        
        c_compiler_autodetected = self.c_compiler is None
        cpp_compiler_autodetected = self.cpp_compiler is None
        as_compiler_autodetected = self.as_compiler is None
        ld_autodetected = self.linker is None
        shared_ld_autodetected = self.shared_linker is None

        toolchain = auto_toolchain(preference, self.c_compiler, self.cpp_compiler, self.as_compiler, self.asm_compiler, self.archiver, self.linker, self.shared_linker)

        if self.c_compiler is None:
            if self.target_is_linux():
                self.c_compiler = T.cast(T.Union[Compiler, None], find_tool(GenericCompiler, toolchain[0], "gcc", "clang"))
            elif self.target_is_windows():
                self.c_compiler = T.cast(T.Union[Compiler, None], find_tool(GenericCompiler, toolchain[0], "msvc", "gcc", "clang-cl", "clang"))

        if self.cpp_compiler is None:
            if self.target_is_linux():
                self.cpp_compiler = T.cast(T.Union[Compiler, None], find_tool(GenericCompiler, toolchain[1], "g++", "clang++"))
            elif self.target_is_windows():
                self.cpp_compiler = T.cast(T.Union[Compiler, None], find_tool(GenericCompiler, toolchain[1], "msvc", "g++", "clang-cl", "clang"))

        if self.as_compiler is None:
            self.as_compiler = T.cast(T.Union[Compiler, None], find_tool(GenericCompiler, toolchain[2], "gcc", "clang"))

        if self.asm_compiler is None:
            self.asm_compiler = T.cast(T.Union[Compiler, None], find_tool(GenericCompiler, toolchain[3], "nasm"))

        if self.archiver is None:
            if self.target_is_linux():
                self.archiver = T.cast(T.Union[Archiver, None], find_tool(GenericArchiver, toolchain[4], "ar", "llvm-ar"))
            elif self.target_is_windows():
                self.archiver = T.cast(T.Union[Archiver, None], find_tool(GenericArchiver, toolchain[4], "msvc", "ar", "llvm-ar"))

        if self.linker is None:
            if self.target_is_linux():
                self.linker = T.cast(T.Union[Linker, None], find_tool(GenericLinker, toolchain[5], "g++", "clang++", "gcc", "clang"))
            elif self.target_is_windows():
                self.linker = T.cast(T.Union[Linker, None], find_tool(GenericLinker, toolchain[5], "msvc", "g++", "clang++", "clang-cl", "gcc", "clang"))

        if self.shared_linker is None:
            if self.target_is_linux():
                self.shared_linker = T.cast(T.Union[SharedLinker, None], find_tool(GenericSharedLinker, toolchain[6], "g++", "clang++", "gcc", "clang"))
            elif self.target_is_windows():
                self.shared_linker = T.cast(T.Union[SharedLinker, None], find_tool(GenericSharedLinker, toolchain[6], "msvc", "g++", "clang++", "clang-cl", "gcc", "clang"))
        
        cc_env = os.getenv("CC")
        if cc_env is not None:
            c_compiler_autodetected = False
            if self.c_compiler is not None:
                os.environ["CCC_CC"] = self.c_compiler.path
                self.c_compiler.reload(cc_env)  # We change the path, but we keep the compiler object
            else:
                self.c_compiler = CompilerGNU(cc_env)
            print_debug_info("Using CC environment variable instead of the config", verbosity)

        cxx_env = os.getenv("CXX")
        if cxx_env is not None:
            cpp_compiler_autodetected = False
            if self.cpp_compiler is not None:
                os.environ["CCC_CXX"] = self.cpp_compiler.path
                self.cpp_compiler.reload(cxx_env)  # We change the path, but we keep the compiler object
            else:
                self.cpp_compiler = CompilerGNUPlusPlus(cxx_env)
            print_debug_info("Using CXX environment variable instead of the config", verbosity)
        
        ld_env = os.getenv("LD")
        if ld_env is not None:
            ld_autodetected = False
            if self.linker is not None:
                self.linker.reload(ld_env)  # We change the path, but we keep the compiler object
            else:
                self.linker = LinkerGNU(ld_env)
            print_debug_info("Using LD environment variable instead of the config", verbosity)
        

        toolchain_tuple = get_toolchain_tuple(
            not c_compiler_autodetected and self.c_compiler is not None and self.c_compiler.path
            or not cpp_compiler_autodetected and self.cpp_compiler is not None and self.cpp_compiler.path
            or not as_compiler_autodetected and self.as_compiler is not None and self.as_compiler.path
            or not ld_autodetected and self.linker is not None and self.linker.path
            or not shared_ld_autodetected and self.shared_linker is not None and self.shared_linker.path
        )
        if toolchain_tuple is not None:
            t, toolchain_prefix = toolchain_tuple
            if c_compiler_autodetected:
                if t == "gcc":
                    compiler = T.cast(T.Callable[[str], Compiler], GenericCompiler("gcc"))(toolchain_prefix + "gcc")
                elif t == "clang":
                    compiler = T.cast(T.Callable[[str], Compiler], GenericCompiler("clang"))(toolchain_prefix + "clang")
                else:
                    compiler = T.cast(T.Callable[[str], Compiler], GenericCompiler("gnu"))(toolchain_prefix + "cc")
                if compiler.is_available():
                    self.c_compiler = compiler
            if cpp_compiler_autodetected:
                if t == "gcc":
                    compiler = T.cast(T.Callable[[str], Compiler], GenericCompiler("g++"))(toolchain_prefix + "g++")
                elif t == "clang":
                    compiler = T.cast(T.Callable[[str], Compiler], GenericCompiler("clang++"))(toolchain_prefix + "clang++")
                else:
                    compiler = T.cast(T.Callable[[str], Compiler], GenericCompiler("gnu++"))(toolchain_prefix + "c++")
                if compiler.is_available():
                    self.cpp_compiler = compiler
            if as_compiler_autodetected:
                if t == "gcc":
                    compiler = T.cast(T.Callable[[str], Compiler], GenericCompiler("gcc"))(toolchain_prefix + "gcc")
                elif t == "clang":
                    compiler = T.cast(T.Callable[[str], Compiler], GenericCompiler("clang"))(toolchain_prefix + "clang")
                else:
                    compiler = T.cast(T.Callable[[str], Compiler], GenericCompiler("gnu"))(toolchain_prefix + "cc")
                if compiler.is_available():
                    self.as_compiler = compiler
            if ld_autodetected:
                if t == "gcc":
                    ld = T.cast(T.Callable[[str], Linker], GenericLinker("g++"))(toolchain_prefix + "g++")
                elif t == "clang":
                    ld = T.cast(T.Callable[[str], Linker], GenericLinker("clang++"))(toolchain_prefix + "clang++")
                else:
                    ld = T.cast(T.Callable[[str], Linker], GenericLinker("gnu++"))(toolchain_prefix + "c++")
                if ld.is_available():
                    self.linker = ld
            if shared_ld_autodetected:
                if t == "gcc":
                    shared_ld = T.cast(T.Callable[[str], SharedLinker], GenericSharedLinker("g++"))(toolchain_prefix + "g++")
                elif t == "clang":
                    shared_ld = T.cast(T.Callable[[str], SharedLinker], GenericSharedLinker("clang++"))(toolchain_prefix + "clang++")
                else:
                    shared_ld = T.cast(T.Callable[[str], SharedLinker], GenericSharedLinker("gnu++"))(toolchain_prefix + "c++")
                if shared_ld.is_available():
                    self.shared_linker = shared_ld
        
        if target_os_autodetected and self.c_compiler is not None and "mingw" in self.c_compiler.path.lower():
            self.target_operating_system = "Windows"

        self.set_debug(self.debug, True)
        self.add_c_cpp_as_asm_flags("-fdiagnostics-color")

    @property
    def c_cpp_flags(self) -> T.List[str]:
        return self.c_flags + self.cpp_flags

    @property
    def c_cpp_as_asm_flags(self) -> T.List[str]:
        return self.c_flags + self.cpp_flags + self.as_flags + self.asm_flags

    def reload_env(self) -> None:
        if self.target_architecture == "" or self.host_architecture == "":
            raise RuntimeError("Unable to load environment because architecture is undetermined")

        self.target_simplified_architecture = simplify_architecture(self.target_architecture)
        self.host_simplified_architecture = simplify_architecture(self.host_architecture)

        if self.target_simplified_architecture == "":
            self.target_simplified_architecture = self.target_architecture
        if self.host_simplified_architecture == "":
            self.host_simplified_architecture = self.host_architecture

        if self.target_is_windows():
            env = load_msvc_environment(os.path.join(self.global_config_dir, "msvc_envs.json"), architecture=self.target_simplified_architecture)
            if env is not None:
                for var in env:
                    os.environ[var] = env[var]

    def reload_tools(self) -> None:
        for tool in (self.c_compiler, self.cpp_compiler, self.as_compiler, self.asm_compiler, self.archiver, self.linker, self.shared_linker):
            if tool is not None:
                tool.reload()
        self.reload_env()

    def reset_build_directories(self) -> None:
        if self.debug:
            mode = "debug"
            old_mode = "release"
        else:
            mode = "release"
            old_mode = "debug"

        if self.obj_build_directory == "":
            self.obj_build_directory = os.path.join("build/.objs/", self.target_operating_system, self.target_simplified_architecture, mode)
        else:
            self.obj_build_directory = replace_architecture(self.obj_build_directory.replace(old_mode, mode), self.target_simplified_architecture)

        if self.exe_build_directory == "":
            self.exe_build_directory = os.path.join("build", self.target_operating_system, self.target_simplified_architecture, mode, "bin")
        else:
            self.exe_build_directory = replace_architecture(self.exe_build_directory.replace(old_mode, mode), self.target_simplified_architecture)

        if self.lib_build_directory == "":
            self.lib_build_directory = os.path.join("build", self.target_operating_system, self.target_simplified_architecture, mode, "lib")
        else:
            self.lib_build_directory = replace_architecture(self.lib_build_directory.replace(old_mode, mode), self.target_simplified_architecture)

    def export_json(self) -> T.Dict[str, T.Any]:
        output = {}
        if self.c_compiler is not None:
            output["c_compiler"] = {
                "type": self.c_compiler.type,
                "path": self.c_compiler.path
            }
        return output

    def copy(self) -> T.Any:
        return copy.deepcopy(self)

    def empty_copy(self, local_config: T.Union[str, None] = None) -> T.Any:
        """Generate a new fresh config object without anything inside. By default, even the local config file isn't used.  
        It can be very helpful if you have a local config file specifying a cross compiler but you want to have the default compiler at some point during the compilation step.
        """
        return Config(self.target_name, args_parsed=self._args_parsed, debug=self.debug, rebuild=self.rebuild, verbosity=self.verbosity, nb_jobs=self.nb_jobs, single_file=self.single_file, compile_commands_dir=self.compile_commands_dir, local_config=local_config)

    def set_debug(self, debug: bool = True, reset_optimization: bool = False) -> None:
        self.debug = debug

        self.reset_build_directories()

        if self.debug:
            self.add_defines("DEBUG")
            self.remove_defines("NDEBUG")
            self.add_c_cpp_as_asm_flags("-g")
            if reset_optimization:
                self.set_optimization("-Og")
        else:
            self.add_defines("NDEBUG")
            self.remove_defines("DEBUG")
            self.remove_c_cpp_as_asm_flags("-g")
            if reset_optimization:
                self.set_optimization("-O3")

    def set_target_architecture(self, architecture: str, reload_tools_and_build_dir: bool = True) -> None:
        self.target_architecture = architecture
        if reload_tools_and_build_dir:
            self.reload_tools()
            self.reset_build_directories()
        else:
            self.reload_env()

        if self.target_simplified_architecture != self.host_simplified_architecture:
            if self.target_simplified_architecture in ("x86", "x64"):
                if self.target_simplified_architecture == "x86":
                    add = "32"
                    remove = "64"
                else:
                    add = "64"
                    remove = "32"
                self.add_c_cpp_flags("-m" + add)
                self.add_ld_flags("-m" + add)
                self.add_shared_linker_flags("-m" + add)
                self.add_as_flags("-m" + add)
                self.remove_c_cpp_flags("-m" + remove)
                self.remove_ld_flags("-m" + remove)
                self.remove_shared_linker_flags("-m" + remove)
                self.remove_as_flags("-m" + remove)
                if self.target_is_windows():
                    self.add_asm_flags("-fwin" + add)
                    self.remove_asm_flags("-fwin" + remove)
                else:
                    self.add_asm_flags("-felf" + add)
                    self.remove_asm_flags("-felf" + remove)

    def set_optimization(self, opt_flag: str) -> None:
        self.remove_c_cpp_as_asm_flags("-O0", "-Og", "-O", "-O1", "-O2", "-O3", "-Os", "-Oz", "-Ofast", "-fomit-frame-pointer")
        self.add_c_cpp_as_asm_flags(opt_flag)

    def get_optimization_level(self) -> T.Union[str, None]:
        for flag in reversed(self.c_cpp_flags):
            if flag in ("-O0", "-Og", "-O", "-O1", "-O2", "-O3", "-Os", "-Oz", "-Ofast", "-fomit-frame-pointer"):
                return flag
        return None

    def target_is_windows(self) -> bool:
        return self.target_operating_system.lower().startswith("win")

    def target_is_linux(self) -> bool:
        return self.target_operating_system.lower().startswith("linux")

    def target_is_mingw(self) -> bool:
        return self.target_is_windows() and isinstance(self.c_compiler, CompilerGNU)

    def add_defines(self, *defines: str) -> None:
        for define in defines:
            if define not in self.defines:
                self.defines.append(define)

    def remove_defines(self, *defines: str) -> None:
        for define in defines:
            if define in self.defines:
                self.defines.remove(define)

    def add_shared_libs(self, *shared_libs: str) -> None:
        for shared_lib in shared_libs:
            if shared_lib not in self.shared_libs:
                self.shared_libs.append(shared_lib)

    def remove_shared_libs(self, *shared_libs: str) -> None:
        for shared_lib in shared_libs:
            if shared_lib in self.shared_libs:
                self.shared_libs.remove(shared_lib)

    def add_includedirs(self, *includedirs: str) -> None:
        for includedir in includedirs:
            if includedir not in self.additional_includedirs:
                self.additional_includedirs.append(includedir)

    def remove_includedirs(self, *includedirs: str) -> None:
        for includedir in includedirs:
            if includedir in self.additional_includedirs:
                self.additional_includedirs.remove(includedir)

    def add_c_flags(self, *c_flags: str) -> None:
        for c_flag in c_flags:
            if c_flag not in self.c_flags:
                self.c_flags.append(c_flag)

    def remove_c_flags(self, *c_flags: str) -> None:
        for c_flag in c_flags:
            if c_flag in self.c_flags:
                self.c_flags.remove(c_flag)
            if self.c_compiler is None:
                continue
            for k in self.c_compiler.translation_dict:
                if c_flag in self.c_compiler.translation_dict[k]:
                    self.c_compiler.translation_dict[k].remove(c_flag)

    def add_cpp_flags(self, *cpp_flags: str) -> None:
        for cpp_flag in cpp_flags:
            if cpp_flag not in self.cpp_flags:
                self.cpp_flags.append(cpp_flag)

    def remove_cpp_flags(self, *cpp_flags: str) -> None:
        for cpp_flag in cpp_flags:
            if cpp_flag in self.cpp_flags:
                self.cpp_flags.remove(cpp_flag)
            if self.cpp_compiler is None:
                continue
            for k in self.cpp_compiler.translation_dict:
                if cpp_flag in self.cpp_compiler.translation_dict[k]:
                    self.cpp_compiler.translation_dict[k].remove(cpp_flag)

    def add_c_cpp_flags(self, *c_cpp_flags: str) -> None:
        self.add_c_flags(*c_cpp_flags)
        self.add_cpp_flags(*c_cpp_flags)

    def remove_c_cpp_flags(self, *c_cpp_flags: str) -> None:
        self.remove_c_flags(*c_cpp_flags)
        self.remove_cpp_flags(*c_cpp_flags)

    def add_c_cpp_as_asm_flags(self, *c_cpp_as_asm_flags: str) -> None:
        self.add_c_flags(*c_cpp_as_asm_flags)
        self.add_cpp_flags(*c_cpp_as_asm_flags)
        self.add_as_flags(*c_cpp_as_asm_flags)
        self.add_asm_flags(*c_cpp_as_asm_flags)

    def remove_c_cpp_as_asm_flags(self, *c_cpp_as_asm_flags: str) -> None:
        self.remove_c_flags(*c_cpp_as_asm_flags)
        self.remove_cpp_flags(*c_cpp_as_asm_flags)
        self.remove_as_flags(*c_cpp_as_asm_flags)
        self.remove_asm_flags(*c_cpp_as_asm_flags)

    def add_as_flags(self, *as_flags: str) -> None:
        for as_flag in as_flags:
            if as_flag not in self.as_flags:
                self.as_flags.append(as_flag)

    def remove_as_flags(self, *as_flags: str) -> None:
        for as_flag in as_flags:
            if as_flag in self.as_flags:
                self.as_flags.remove(as_flag)
            if self.as_compiler is None:
                continue
            for k in self.as_compiler.translation_dict:
                if as_flag in self.as_compiler.translation_dict[k]:
                    self.as_compiler.translation_dict[k].remove(as_flag)

    def add_asm_flags(self, *asm_flags: str) -> None:
        for asm_flag in asm_flags:
            if asm_flag not in self.asm_flags:
                self.asm_flags.append(asm_flag)

    def remove_asm_flags(self, *asm_flags: str) -> None:
        for asm_flag in asm_flags:
            if asm_flag in self.asm_flags:
                self.asm_flags.remove(asm_flag)
            if self.asm_compiler is None:
                continue
            for k in self.asm_compiler.translation_dict:
                if asm_flag in self.asm_compiler.translation_dict[k]:
                    self.asm_compiler.translation_dict[k].remove(asm_flag)

    def add_ar_flags(self, *ar_flags: str) -> None:
        for ar_flag in ar_flags:
            if ar_flag not in self.ar_flags:
                self.ar_flags.append(ar_flag)

    def remove_ar_flags(self, *ar_flags: str) -> None:
        for ar_flag in ar_flags:
            if ar_flag in self.ar_flags:
                self.ar_flags.remove(ar_flag)
            if self.archiver is None:
                continue
            for k in self.archiver.translation_dict:
                if ar_flag in self.archiver.translation_dict[k]:
                    self.archiver.translation_dict[k].remove(ar_flag)

    def add_ld_flags(self, *ld_flags: str) -> None:
        for ld_flag in ld_flags:
            if ld_flag not in self.ld_flags:
                self.ld_flags.append(ld_flag)

    def remove_ld_flags(self, *ld_flags: str) -> None:
        for ld_flag in ld_flags:
            if ld_flag in self.ld_flags:
                self.ld_flags.remove(ld_flag)
            if self.linker is None:
                continue
            for k in self.linker.translation_dict:
                if ld_flag in self.linker.translation_dict[k]:
                    self.linker.translation_dict[k].remove(ld_flag)

    def add_shared_linker_flags(self, *shared_linker_flags: str) -> None:
        for shared_linker_flag in shared_linker_flags:
            if shared_linker_flag not in self.shared_linker_flags:
                self.shared_linker_flags.append(shared_linker_flag)

    def remove_shared_linker_flags(self, *shared_linker_flags: str) -> None:
        for shared_linker_flag in shared_linker_flags:
            if shared_linker_flag in self.shared_linker_flags:
                self.shared_linker_flags.remove(shared_linker_flag)
            if self.shared_linker is None:
                continue
            for k in self.shared_linker.translation_dict:
                if shared_linker_flag in self.shared_linker.translation_dict[k]:
                    self.shared_linker.translation_dict[k].remove(shared_linker_flag)

    def add_exported_headers(self, *exported_headers: str, subfolder: T.Union[str, None] = None) -> None:
        for exported_header in exported_headers:
            if (exported_header, subfolder) not in self.exported_headers:
                self.exported_headers.append((exported_header, subfolder))

    def remove_exported_headers(self, *exported_headers: str, subfolder: T.Union[str, None] = None) -> None:
        for exported_header in exported_headers:
            if (exported_header, subfolder) in self.exported_headers:
                self.exported_headers.remove((exported_header, subfolder))



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import typing as T

from .common import Compiler


_powermake_warning_flags_to_nasm_flags: T.Dict[str, T.List[str]] = {
    "-Wextra": ["-Wall"],
    "-Weverything": ["-Wall"],
    "-Wsecurity": ["-Wall"],
    "-Wpedantic": [],
    "-pedantic": [],
    "-Wswitch": [],
    "-Wswitch-enum": [],
    "-fanalyzer": ["-Wall"],
    "-ffuzzer": []
}

_powermake_optimization_flags_to_nasm_flags: T.Dict[str, T.List[str]] = {
    "-Og": ["-O1"],
    "-O2": ["-Ox"],
    "-O3": ["-Ox"],
    "-Os": ["-Ox"],
    "-Oz": ["-Ox"],
    "-Ofast": ["-Ox"],
    "-fomit-frame-pointer": [],
}

_powermake_machine_flags_to_nasm_flags: T.Dict[str, T.List[str]] = {
    "-m32": ["-felf32"],
    "-m64": ["-felf64"],
    "-mmmx": [],
    "-msse": [],
    "-msse2": [],
    "-mssse3": [],
    "-mavx": [],
    "-mavx2": []
}

# These flags doesn't inherit from tools._powermake_flags_to_gnu_flags because 
_powermake_flags_to_nasm_flags: T.Dict[str, T.List[str]] = {
    **_powermake_warning_flags_to_nasm_flags,
    **_powermake_optimization_flags_to_nasm_flags,
    **_powermake_machine_flags_to_nasm_flags,
    "-fdiagnostics-color": [],
    "-fsecurity=1": [],
    "-fsecurity=2": [],
    "-fsecurity": []
}


class CompilerNASM(Compiler):
    type: T.ClassVar = "nasm"
    default_path: T.ClassVar = "nasm"
    obj_extension: T.ClassVar = ".o"
    translation_dict: T.ClassVar = _powermake_flags_to_nasm_flags

    def __init__(self, path: str = "nasm"):
        super().__init__(path)

    def format_args(self, defines: T.List[str], includedirs: T.List[str], flags: T.List[str] = []) -> T.List[str]:
        return [f"-d{define}" for define in defines] + [f"-i{includedir}" for includedir in includedirs] + self.translate_flags(flags)

    def basic_compile_command(self, outputfile: str, inputfile: str, args: T.List[str] = []) -> T.List[str]:
        return [self.path, "-o", outputfile, inputfile, *args]

    def check_if_arg_exists(self, arg: str) -> bool:
        return subprocess.run([self.path, arg, "-E", "-x", "assemble", "/dev/null"], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL).returncode == 0



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "Linker", "LinkerGNU", "LinkerLD", "LinkerGCC", "LinkerClang", "LinkerGPlusPlus", "LinkerClangPlusPlus", "LinkerMSVC", "LinkerClang_CL"
]

import typing as T

from .common import Linker
from .gnu import LinkerGNU, LinkerLD, LinkerGCC, LinkerClang, LinkerGPlusPlus, LinkerClangPlusPlus
from .msvc import LinkerMSVC, LinkerClang_CL


_linker_types: T.Dict[str, T.Callable[[], Linker]] = {
    "gnu": LinkerGNU,
    "ld": LinkerLD,
    "gcc": LinkerGCC,
    "g++": LinkerGPlusPlus,
    "clang": LinkerClang,
    "clang++": LinkerClangPlusPlus,
    "msvc": LinkerMSVC,
    "clang-cl": LinkerClang_CL
}


def GenericLinker(linker_type: str) -> T.Union[T.Callable[[], Linker], None]:
    if linker_type not in _linker_types:
        return None
    return _linker_types[linker_type]


def get_all_linker_types() -> T.Set[str]:
    return set(_linker_types.keys())



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import shutil
import argparse
import typing as T

from .config import Config
from .utils import makedirs
from .__version__ import __version__
from .interactive_config import InteractiveConfig
from .display import print_info, print_debug_info, init_colors


def default_on_clean(config: Config) -> None:
    """Remove exe, lib and obj directories

    Args:
        config (Config): the object given by `powermake.run`
    """
    print_info("Cleaning", config.verbosity)

    if config.exe_build_directory is not None and os.path.isdir(config.exe_build_directory):
        print_debug_info(f"Removing {config.exe_build_directory}", config.verbosity)
        shutil.rmtree(config.exe_build_directory)

    if config.lib_build_directory is not None and os.path.isdir(config.lib_build_directory):
        print_debug_info(f"Removing {config.lib_build_directory}", config.verbosity)
        shutil.rmtree(config.lib_build_directory)

    if config.obj_build_directory is not None and os.path.isdir(config.obj_build_directory):
        print_debug_info(f"Removing {config.obj_build_directory}", config.verbosity)
        shutil.rmtree(config.obj_build_directory)


def default_on_install(config: Config, location: T.Union[str, None]) -> None:
    """Install binary files, library files and exported headers into the subfolders `bin/`, `lib/` and `include/` of the location provided.

    You can directly give this function as a callback to `powermake.run` (it's the default behavior), but it's even better to call it from your own implementation of the install_callback.
    Like that, you can do work before this function, especially adding exported headers to the config with `config.add_exported_headers`

    If `location` is None, his default value is `./install/`

    For each exported header, if it has been exported with a None subfolder, it is simply copied into `include/`. If it has been exported with a subfolder, it is copied into `include/{subfolder}/`

    Args:
        config (Config): The powermake.Config object created by the `powermake.run`. You should have called the method `config.add_exported_headers` on this object before giving it to this function.
        location (str): Where to install `bin/`, `lib/` and `include/` subfolders. Ideally, if the location given to the install_callback by `powermake.run` was not None, you should give that to this function.
    """
    if location is None:
        location = "./install/"

    print_info(f"Installing to {location}", config.verbosity)

    nb_files_installed = 0

    lib_folder = os.path.join(location, "lib")
    include_folder = os.path.join(location, "include")
    bin_folder = os.path.join(location, "bin")

    if os.path.isdir(config.lib_build_directory):
        lib_files = os.listdir(config.lib_build_directory)
        if lib_files != []:
            makedirs(lib_folder, exist_ok=True)
            for file in lib_files:
                src = os.path.join(config.lib_build_directory, file)
                dest = os.path.join(lib_folder, file)
                print_debug_info(f"copying {src} to {dest}", config.verbosity)
                nb_files_installed += 1
                shutil.copy(src, dest)

    if os.path.isdir(config.exe_build_directory):
        bin_files = os.listdir(config.exe_build_directory)
        if bin_files != []:
            makedirs(bin_folder, exist_ok=True)
            for file in bin_files:
                src = os.path.join(config.exe_build_directory, file)
                dest = os.path.join(bin_folder, file)
                print_debug_info(f"copying {src} to {dest}", config.verbosity)
                nb_files_installed += 1
                shutil.copy(src, dest)

    if len(config.exported_headers) != 0:
        makedirs(include_folder, exist_ok=True)
    for (src, subfolder) in config.exported_headers:
        if subfolder is None:
            dest = os.path.join(include_folder, os.path.basename(src))
        else:
            makedirs(os.path.join(include_folder, subfolder), exist_ok=True)
            dest = os.path.join(include_folder, subfolder, os.path.basename(src))
        print_debug_info(f"copying {src} to {dest}", config.verbosity)
        nb_files_installed += 1
        shutil.copy(src, dest)

    print_info(f"{nb_files_installed} files successfully copied", config.verbosity)

def get_version_str() -> str:
    return f"PowerMake {__version__}\nCopyright (C) 2024 Macéo Tuloup\nLicense Apache-2.0: <http://www.apache.org/licenses/LICENSE-2.0>\nThis is a free software; see the sources for copying conditions: <https://github.com/mactul/powermake>\nThere is NO WARRANTY, to the extent permitted by law."

class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, prog: T.Union[str, None] = None, description: T.Union[str, None] = None, **kwargs: T.Any):
        if prog is None:
            prog = "python YOUR_MAKEFILE.py"
        if description is None:
            description = f"PowerMake {__version__}"
        super().__init__(prog=prog, description=description, **kwargs)

        self.add_argument("action", choices=["build", "clean", "install", "config"], nargs='?')
        self.add_argument("install_location", nargs='?', help="Only if the action is set to install, indicate in which folder the installation should be")
        self.add_argument("-d", "--debug", help="Trigger the build function with config.debug set to True.", action="store_true")
        self.add_argument("-b", "--build", help="Trigger the build function. This is the default but it can be used in combination with --clean or --install", action="store_true")
        self.add_argument("-r", "--rebuild", help="Trigger the build function with config.rebuild set to True.", action="store_true")
        self.add_argument("-c", "--clean", help="Trigger the clean function.", action="store_true")
        self.add_argument("-i", help="Trigger the install function with the location argument set to None.", action="store_true")
        self.add_argument("--install", nargs='?', metavar="LOCATION", help="Trigger the install function with the location argument set to the location given or None.", default=False)
        self.add_argument("-f", "--config", help="Switch to an interactive mode that helps you editing your configuration files.", action="store_true")
        self.add_argument("-q", "--quiet", help="Disable all messages from the lib.", action="store_true")
        self.add_argument("-v", "--verbose", help="Display every command the lib runs.", action="store_true")
        self.add_argument("-j", "--jobs", help="Set on how many threads the compilation should be parallelized.", default=0, type=int)
        self.add_argument("-l", "--local-config", metavar="LOCAL_CONFIG_PATH", help="Set the path for the local config", default="./powermake_config.json")
        self.add_argument("-g", "--global-config", metavar="GLOBAL_CONFIG_PATH", help="Set the path for the global config", default=None)
        self.add_argument("-s", "--single-file", metavar="FILE", help="Run the compilation but only compile the specified file.", default=None)
        self.add_argument("-o", "--compile-commands-dir", metavar="DIRECTORY", help="Run the compilation and generate a compile_commands.json file in the directory specified.", default=None)
        self.add_argument("--get-lib-build-folder", help="(Internal option) - Return the lib build folder path according to the config.", action="store_true")
        self.add_argument("--retransmit-colors", help="(Internal option) - Let all ANSI color codes intact, even if not in a terminal.", action="store_true")
        self.add_argument("--version", help="display PowerMake version", action="store_true")


def generate_config(target_name: str, args_parsed: T.Union[argparse.Namespace, None] = None) -> Config:
    """Parse the command line and create a powermake.Config object according to the command line arguments."""
    parser = None
    if args_parsed is None:
        parser = ArgumentParser()
        args_parsed = parser.parse_args()

    if args_parsed.install_location is not None and args_parsed.action != "install":
        print(f"Unexpected argument {args_parsed.install_location} with an action different from 'install'", file=sys.stderr)
        if parser is not None:
            parser.print_usage(file=sys.stderr)
        exit(1)

    if args_parsed.quiet and args_parsed.verbose:
        print("Passing --quiet and --verbose arguments in the same time doesn't make any sense.", file=sys.stderr)
        if parser is not None:
            parser.print_usage(file=sys.stderr)
        exit(1)
    
    if args_parsed.version:
        print(get_version_str())
        exit(0)

    if args_parsed.quiet:
        verbosity = 0
    elif args_parsed.verbose:
        verbosity = 2
    else:
        verbosity = 1

    if not args_parsed.retransmit_colors:
        init_colors()

    if args_parsed.action == "config" or args_parsed.config:
        InteractiveConfig(global_config=args_parsed.global_config, local_config=args_parsed.local_config)

    config = Config(target_name, args_parsed=args_parsed, verbosity=verbosity, debug=args_parsed.debug, rebuild=args_parsed.rebuild, local_config=args_parsed.local_config, global_config=args_parsed.global_config, nb_jobs=args_parsed.jobs, single_file=args_parsed.single_file, compile_commands_dir=args_parsed.compile_commands_dir)

    return config


def run_callbacks(config: Config, *, build_callback: T.Callable[[Config], None], clean_callback: T.Callable[[Config], None] = default_on_clean, install_callback: T.Callable[[Config, T.Union[str, None]], None] = default_on_install) -> None:
    """From a config generated by `powermake.generate_config` start each callback according to the command line."""

    clean = False
    build = False
    install = False
    interactive_config = False
    if config._args_parsed is None:
        raise RuntimeError("powermake.run_callbacks should only be run with a config generated by powermake.generate_config")

    if config._args_parsed.action == "clean" or config._args_parsed.clean:
        clean = True
    if config._args_parsed.action == "build" or config._args_parsed.build or config._args_parsed.rebuild:
        build = True
    if config._args_parsed.action == "install" or config._args_parsed.install is not False or config._args_parsed.i:
        install = True
    if config._args_parsed.action == "config" or config._args_parsed.config:
        interactive_config = True

    if clean:
        clean_callback(config)
    if build or (not clean and not install and not interactive_config):
        build_callback(config)
    if install:
        if config._args_parsed.action == "install":
            install_callback(config, config._args_parsed.install_location)
        else:
            install_callback(config, config._args_parsed.install or None)

    # After doing all compilation, if the argument get-lib-build-folder was given, we print the absolute path of the directory in which all lib have been built and exit.
    # Like that, the last line of the program output will be the requested folder.
    # This is used by the powermake.run_another_powermake function.
    if config._args_parsed.get_lib_build_folder:
        if config.lib_build_directory is not None and os.path.exists(config.lib_build_directory):
            print(os.path.abspath(config.lib_build_directory))
        else:
            print()
        exit(0)


def run(target_name: str, *, build_callback: T.Callable[[Config], None], clean_callback: T.Callable[[Config], None] = default_on_clean, install_callback: T.Callable[[Config, T.Union[str, None]], None] = default_on_install, args_parsed: T.Union[argparse.Namespace, None] = None) -> None:
    """Parse the command line, create a powermake.Config object according to the command line arguments and call the appropriate callback.

    If multiples actions are passed trough the command line, like `-cri` (or `--clean --rebuild --install` in the long form), multiple callbacks can be called.  
    Callbacks are always called in this order: clean -> build -> install.

    Warning: You shouldn't run anything important after the return of this function, because if `--get-lib-build-folder` is provided on the command line, this function call `exit(0)`

    Args:
        target_name (str): The name of your application. It will be stored in the config and used for default naming.
        build_callback (callable): The function that will be called when building or rebuilding. This callback takes a single parameter: config
        clean_callback (callable, optional): The function that will be called when cleaning. This callback takes a single parameter: config
        install_callback (callable, optional): The function that will be called when installing. This callback takes 2 parameters: config and location.  
            If the location is not provided on the command line, his value is None
        args_parsed (argparse.Namespace, optional): Should be left to None. If set, it should be created by powermake.ArgumentParser().parse_args() to have all the required members for the Namespace
    """

    config = generate_config(target_name, args_parsed)

    run_callbacks(config, build_callback=build_callback, clean_callback=clean_callback, install_callback=install_callback)



import os
import tempfile


def makedirs(path: str, exist_ok: bool = True) -> None:
    if path != '':
        os.makedirs(path, exist_ok=exist_ok)


_tempdir = None
_empty_file = None
def get_empty_file() -> str:
    global _empty_file
    global _tempdir
    if os.path.exists("/dev/null"):
        return "/dev/null"
    if _empty_file is None or not os.path.exists(_empty_file):
        _tempdir = tempfile.TemporaryDirectory("powermake_utils")
        _empty_file = os.path.join(_tempdir.name, "emptyfile")
        file = open(_empty_file, "w")
        file.close()
    
    return _empty_file
    


# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "Compiler", "CompilerNASM", "CompilerMSVC", "CompilerClang_CL", "CompilerGNU",
    "CompilerGNUPlusPlus", "CompilerGCC", "CompilerGPlusPlus", "CompilerClang", "CompilerClangPlusPlus"
]

import typing as T

from .common import Compiler
from .asm import CompilerNASM
from .msvc import CompilerMSVC, CompilerClang_CL
from .gnu import CompilerGNU, CompilerGNUPlusPlus, CompilerGCC, CompilerGPlusPlus, CompilerClang, CompilerClangPlusPlus


_c_compiler_types: T.Dict[str, T.Callable[[], Compiler]] = {
    "gnu": CompilerGNU,
    "gcc": CompilerGCC,
    "clang": CompilerClang,
    "msvc": CompilerMSVC,
    "clang-cl": CompilerClang_CL
}

_cpp_compiler_types: T.Dict[str, T.Callable[[], Compiler]] = {
    "gnu++": CompilerGNUPlusPlus,
    "g++": CompilerGPlusPlus,
    "clang++": CompilerClangPlusPlus,
    "msvc": CompilerMSVC,
    "clang-cl": CompilerClang_CL
}


_as_compiler_types: T.Dict[str, T.Callable[[], Compiler]] = {
    "gnu": CompilerGNU,
    "gcc": CompilerGCC,
    "clang": CompilerClang
}

_asm_compiler_types: T.Dict[str, T.Callable[[], Compiler]] = {
    "nasm": CompilerNASM
}

_compiler_types: T.Dict[str, T.Callable[[], Compiler]] = {
    **_c_compiler_types,
    **_cpp_compiler_types,
    **_as_compiler_types,
    **_asm_compiler_types
}


def GenericCompiler(compiler_type: str) -> T.Union[T.Callable[[], Compiler], None]:
    if compiler_type not in _compiler_types:
        return None
    return _compiler_types[compiler_type]


def get_all_compiler_types() -> T.Set[str]:
    return set(_compiler_types.keys())


def get_all_c_compiler_types() -> T.Set[str]:
    return set(_c_compiler_types.keys())


def get_all_cpp_compiler_types() -> T.Set[str]:
    return set(_cpp_compiler_types.keys())


def get_all_as_compiler_types() -> T.Set[str]:
    return set(_as_compiler_types.keys())


def get_all_asm_compiler_types() -> T.Set[str]:
    return set(_asm_compiler_types.keys())



# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .args_parser import generate_config, get_version_str

def main() -> None:
    generate_config(target_name="NULL")  # just to read the command line

    print(get_version_str())

if __name__ == "__main__":
    main()


# Copyright 2024 Macéo Tuloup

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys
import glob
import json
import fnmatch
import subprocess
import importlib.util
import __main__ as __makefile__
import typing as T
from concurrent.futures import ThreadPoolExecutor

from .config import Config
from .utils import makedirs
from .__version__ import __version__
from .display import print_info, print_debug_info
from .operation import Operation, needs_update, run_command
from .args_parser import run, default_on_clean, default_on_install, ArgumentParser, generate_config, run_callbacks


if hasattr(__makefile__, '__file__'):
    os.chdir(os.path.dirname(os.path.realpath(__makefile__.__file__)))


def import_module(module_name: str, module_path: T.Union[str, None] = None) -> T.Any:
    """Import a custom module from a path

    Args:
        module_name (str): The name of the module once it will be imported
        module_path (str, optional): The path of the module, if None, it takes the module_name as a path.

    Returns:
        module: an module object, that you can use as a namespace
    """
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    if spec is None or spec.loader is None:
        return None
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def get_files(*patterns: str) -> T.Set[str]:
    """Return all files on the disk that matches one of the patterns given

    Supported patterns are `*`, like `./*.c` and `**/` like `./**/*.c` to search all .c recursively, starting at `./`

    Warning: `**.c` will not return all .c recursively, you have to use `**/*.c` for that.

    Args:
        patterns (str): As many patterns as you want to include

    Returns:
        set: A set (unordered list) of files. You can use the methods `add` and `update` to add files to this set.
    """
    files = set()
    for pattern in patterns:
        files.update(glob.glob(pattern, recursive=True))
    return files


def _match_a_pattern(file: str, patterns: T.Iterable[str]) -> bool:
    for pattern in patterns:
        if fnmatch.fnmatch(file, pattern):
            return True
    return False


def filter_files(files: T.Set[str], *patterns: str) -> T.Set[str]:
    """Create a copy of the set `files` with all elements that matches `pattern` removed.

    This function will equally works if `files` is an iterable but will always returns a set.

    Args:
        files (set): The set of files to filter
        patterns (str): As many patterns as you want to exclude

    Returns:
        set: the filtered set
    """
    output = set()
    for file in files:
        if not _match_a_pattern(file, patterns):
            output.add(file)
    return output


def file_in_files_set(file: str, files_set: T.Iterable[str]) -> bool:
    for f in files_set:
        if os.path.samefile(f, file):
            return True
    return False


def compile_files(config: Config, files: T.Union[T.Set[str], T.List[str]], force: T.Union[bool, None] = None) -> T.Union[T.Set[str], T.List[str]]:
    """Compile each C/C++/ASM file in the `files` set according to the compiler and options stored in `config`

    The compilation is parallelized with `config.nb_jobs` threads

    If `files` is a list, compile_files will return a list instead of a set, with the order preserved

    Args:
        config (Config): A powermake.Config object, containing all directives for the compilation. Either the one given to the build_callback or a modified copy.
        files (set): A set of files that ends by .c, .cpp, .cc, .C, .s, .S or .asm to compile in .o (or the specified compiler equivalent extension)
        force (bool, optional): Whether the function should verify if a file needs to be recompiled or if it should recompile everything.  
            If not specified, this parameter takes the value of config.rebuild

    Raises:
        RuntimeError: if no compiler is found
        ValueError: if `files` contains a file that doesn't ends with .c, .cpp, .cc, .C, .s, .S or .asm

    Returns:
        set: A set of object filepaths generated by by the compilation
    """
    generated_objects: T.Union[T.Set[str], T.List[str]] = set()

    operations: T.Union[T.Set[Operation], T.List[Operation]]

    if isinstance(files, set):
        operations = set()
    else:
        operations = []

    if force is None:
        force = config.rebuild

    if config.single_file is not None:
        if file_in_files_set(config.single_file, files):
            if isinstance(files, set):
                files = {config.single_file}
            else:
                files = [config.single_file]
        else:
            if isinstance(files, set):
                return set()
            return []

    if config.c_compiler is not None:
        c_args = config.c_compiler.format_args(config.defines, config.additional_includedirs, config.c_flags)
    else:
        c_args = []

    if config.cpp_compiler is not None:
        cpp_args = config.cpp_compiler.format_args(config.defines, config.additional_includedirs, config.cpp_flags)
    else:
        cpp_args = []

    if config.as_compiler is not None:
        as_args = config.as_compiler.format_args(config.defines, config.additional_includedirs, config.as_flags)
    else:
        as_args = []

    if config.asm_compiler is not None:
        asm_args = config.asm_compiler.format_args(config.defines, config.additional_includedirs, config.asm_flags)
    else:
        asm_args = []

    for file in files:
        if config.c_compiler is not None:
            obj_extension = config.c_compiler.obj_extension
        elif config.cpp_compiler is not None:
            obj_extension = config.cpp_compiler.obj_extension
        else:
            raise RuntimeError("No C/C++ compiler has been specified and the default config didn't find any")

        output_file = os.path.normpath(config.obj_build_directory + "/" + file.replace("..", "__") + obj_extension)
        makedirs(os.path.dirname(output_file), exist_ok=True)

        if "../" in file:
            # if the path of the file contains ../, the ide might not be able to understand warning messages.
            file = os.path.abspath(file)

        if file.endswith(".c"):
            if config.c_compiler is None:
                raise RuntimeError("No C compiler has been specified and the default config didn't find any")
            command = config.c_compiler.basic_compile_command(output_file, file, c_args)
        elif file.endswith((".cpp", ".cc", ".C")):
            if config.cpp_compiler is None:
                raise RuntimeError("No C++ compiler has been specified and the default config didn't find any")
            command = config.cpp_compiler.basic_compile_command(output_file, file, cpp_args)
        elif file.endswith((".s", ".S")):
            if config.as_compiler is None:
                raise RuntimeError("No AS compiler has been specified and the default config didn't find any")
            command = config.as_compiler.basic_compile_command(output_file, file, as_args)
        elif file.endswith(".asm"):
            if config.asm_compiler is None:
                raise RuntimeError("No ASM compiler has been specified and the default config didn't find any")
            command = config.asm_compiler.basic_compile_command(output_file, file, asm_args)
        else:
            raise ValueError("The file extension %s can't be compiled", (os.path.splitext(file)[1], ))
        op = Operation(output_file, {file}, config, command)
        if isinstance(files, set):
            T.cast(T.Set[Operation], operations).add(op)
        else:
            T.cast(T.List[Operation], operations).append(op)

    if config.compile_commands_dir is not None:
        json_commands = []
        for op in operations:
            json_commands.append(op.get_json_command())

        makedirs(config.compile_commands_dir, exist_ok=True)
        with open(os.path.join(config.compile_commands_dir, "compile_commands.json"), "w") as fd:
            json.dump(json_commands, fd, indent=4)

    if config.single_file is not None:
        (op, ) = operations
        op.execute(force)
        exit(0)

    with ThreadPoolExecutor(max_workers=config.nb_jobs) as executor:
        output = executor.map(lambda op: op.execute(force), operations)
        if isinstance(files, set):
            generated_objects = set(output)
        else:
            generated_objects = list(output)

    return generated_objects


def archive_files(config: Config, object_files: T.Iterable[str], archive_name: T.Union[str, None] = None, force: T.Union[bool, None] = None) -> str:
    """Create a static library from A set of object files.

    Args:
        config (Config): A powermake.Config object, containing all directives for the compilation. Either the one given to the build_callback or a modified copy.
        object_files (set or list): A set of object files, potentially the set generated by `compile_files`. It can be a list if the order matters.
        archive_name (str, optional): The name of the static library you want to create, minus the extension. If None, it will be lib{config.target_name}
        force (bool, optional): Whether the function should verify if the static library needs to be re-archived or if it should re-archive in any case.  
            If not specified, this parameter takes the value of config.rebuild

    Raises:
        RuntimeError: if no archiver is found

    Returns:
        str: the path of the archive generated
    """
    if force is None:
        force = config.rebuild

    if archive_name is None:
        archive_name = "lib" + config.target_name

    if config.archiver is None:
        raise RuntimeError("No archiver has been specified and the default config didn't find any")
    output_file = os.path.normpath(config.lib_build_directory + "/" + archive_name + config.archiver.static_lib_extension)
    makedirs(os.path.dirname(output_file), exist_ok=True)
    command = config.archiver.basic_archive_command(output_file, object_files, config.ar_flags)
    return Operation(output_file, object_files, config, command).execute(force=force)


def link_files(config: Config, object_files: T.Iterable[str], archives: T.List[str] = [], executable_name: T.Union[str, None] = None, force: T.Union[bool, None] = None) -> str:
    """Create an executable from a list of object files and a list of archive files.

    Args:
        config (Config): A powermake.Config object, containing all directives for the compilation. Either the one given to the build_callback or a modified copy.
        object_files (set or list): A set of object files, potentially the set generated by `compile_files`. It can be a list if the order matters.
        archives (list): A set of static libraries filepaths
        executable_name (str, optional): The name of the executable you want to create, minus the extension. If None, it will be config.target_name
        force (bool, optional): Whether the function should verify if the executable needs to be re-linked or if it should re-link in any case.  
            If not specified, this parameter takes the value of config.rebuild

    Raises:
        RuntimeError: if no linker is found

    Returns:
        str: the path of the executable generated
    """
    if force is None:
        force = config.rebuild

    if executable_name is None:
        executable_name = config.target_name

    if config.linker is None:
        raise RuntimeError("No linker has been specified and the default config didn't find any")
    extension = ""
    if config.target_is_windows():
        extension = ".exe"
    output_file = os.path.normpath(config.exe_build_directory + "/" + executable_name + extension)
    makedirs(os.path.dirname(output_file), exist_ok=True)
    args = config.linker.format_args(shared_libs=config.shared_libs, flags=config.ld_flags)
    command = config.linker.basic_link_command(output_file, object_files, archives, args)
    return Operation(output_file, set(object_files).union(archives), config, command).execute(force=force)


def link_shared_lib(config: Config, object_files: T.Iterable[str], archives: T.List[str] = [], lib_name: T.Union[str, None] = None, force: T.Union[bool, None] = None) -> str:
    """Create a shared library from a list of object files and a list of archive files.

    Args:
        config (Config): A powermake.Config object, containing all directives for the compilation. Either the one given to the build_callback or a modified copy.
        object_files (set): A set of object files, potentially the set generated by `compile_files`
        archives (list): A set of static libraries filepaths
        lib_name (str, optional): The name of the library you want to create, minus the extension. If None, it will be lib{config.target_name}
        force (bool, optional): Whether the function should verify if the library needs to be re-linked or if it should re-link in any case.  
            If not specified, this parameter takes the value of config.rebuild

    Raises:
        RuntimeError: if no shared linker is found

    Returns:
        str: the path of the library generated
    """
    if force is None:
        force = config.rebuild

    if lib_name is None:
        lib_name = "lib" + config.target_name

    if config.shared_linker is None:
        raise RuntimeError("No shared linker has been specified and the default config didn't find any")
    output_file = os.path.normpath(config.lib_build_directory + "/" + lib_name + config.shared_linker.shared_lib_extension)
    makedirs(os.path.dirname(output_file), exist_ok=True)
    args = config.shared_linker.format_args(shared_libs=config.shared_libs, flags=config.ld_flags)
    command = config.shared_linker.basic_link_command(output_file, object_files, archives, args)
    return Operation(output_file, set(object_files).union(archives), config, command).execute(force=force)


def delete_files_from_disk(*filepaths: str) -> None:
    """Delete a set of filepaths from the disk and ignore them if they don't exists
    """
    for filepath in filepaths:
        try:
            os.remove(filepath)
        except OSError:
            pass


def run_another_powermake(config: Config, path: str, debug: T.Union[bool, None] = None, rebuild: T.Union[bool, None] = None, verbosity: T.Union[int, None] = None, nb_jobs: T.Union[int, None] = None) -> T.Union[T.List[str], None]:
    """Run a powermake from another directory and returns a list of path to all libraries generated

    Args:
        config (Config): A powermake.Config object, containing all directives for the compilation. Either the one given to the build_callback or a modified copy.
        path (str): The path of the powermake to run
        debug (bool, optional): Whether the other powermake should be run in debug mode.  
            If not specified, this parameter takes the value of config.debug
        rebuild (bool, optional): Whether the other powermake should be run in rebuild mode.  
            If not specified, this parameter takes the value of config.rebuild
        verbosity (int, optional): With which verbosity level the other powermake should be run.  
            If not specified, this parameter takes the value of config.verbosity
        nb_jobs (int, optional): With how many threads the other powermake should be run.  
            If not specified, this parameter takes the value of config.nb_jobs

    Raises:
        RuntimeError: if the other powermake fails

    Returns:
        list: A list of path to all libraries generated
    """
    if debug is None:
        debug = config.debug
    if rebuild is None:
        rebuild = config.rebuild
    if verbosity is None:
        verbosity = config.verbosity
    if nb_jobs is None:
        nb_jobs = config.nb_jobs

    command = [sys.executable, path, "--get-lib-build-folder", "--retransmit-colors", "-j", str(nb_jobs)]
    if verbosity == 0:
        command.append("-q")
    elif verbosity >= 2:
        command.append("-v")

    if rebuild:
        command.append("-r")

    if debug:
        command.append("-d")

    print_info(f"Running {path}", config.verbosity)
    print_debug_info(command, config.verbosity)

    try:
        output = subprocess.check_output(command, encoding="utf-8").splitlines()
    except OSError:
        raise RuntimeError(f"Failed to run powermake {path}")

    if verbosity != 0:
        for line in output[:-1]:
            print(line)

    path = output[-1]
    if path != "":
        return [os.path.join(path, file) for file in os.listdir(path)]
    return None
